import re
from dataclasses import dataclass
from typing import Any, Callable, Optional, Pattern, TYPE_CHECKING, Tuple, Union

from apischema.metadata.keys import (
    CONVERSION_METADATA,
    DEFAULT_AS_SET_METADATA,
    FALL_BACK_ON_DEFAULT_METADATA,
    FLATTEN_METADATA,
    INIT_VAR_METADATA,
    NONE_AS_UNDEFINED_METADATA,
    POST_INIT_METADATA,
    PROPERTIES_METADATA,
    REQUIRED_METADATA,
    SKIP_METADATA,
    VALIDATORS_METADATA,
)
from apischema.types import AnyType, Metadata, MetadataImplem, MetadataMixin

if TYPE_CHECKING:
    from apischema.conversions.conversions import AnyConversion
    from apischema.validation.validators import Validator


def simple_metadata(key: str) -> Metadata:
    return MetadataImplem({key: ...})


@dataclass(frozen=True)
class ConversionMetadata(MetadataMixin):
    key = CONVERSION_METADATA
    deserialization: Optional["AnyConversion"] = None
    serialization: Optional["AnyConversion"] = None


conversion = ConversionMetadata

default_as_set = simple_metadata(DEFAULT_AS_SET_METADATA)

fall_back_on_default = simple_metadata(FALL_BACK_ON_DEFAULT_METADATA)

flatten = simple_metadata(FLATTEN_METADATA)
flattened = flatten
merged = flatten


def init_var(tp: AnyType) -> Metadata:
    return MetadataImplem({INIT_VAR_METADATA: tp})


none_as_undefined = simple_metadata(NONE_AS_UNDEFINED_METADATA)

post_init = simple_metadata(POST_INIT_METADATA)


class PropertiesMetadata(dict, Metadata):  # type: ignore
    def __init__(self):
        super().__init__({PROPERTIES_METADATA: None})

    def __call__(
        self, pattern: Union[str, Pattern, "ellipsis"]  # noqa: F821
    ) -> Metadata:
        if pattern is not ...:
            pattern = re.compile(pattern)
        return MetadataImplem({PROPERTIES_METADATA: pattern})


properties = PropertiesMetadata()

required = simple_metadata(REQUIRED_METADATA)


@dataclass(frozen=True)
class SkipMetadata(MetadataMixin):
    key = SKIP_METADATA
    deserialization: bool = False
    serialization: bool = False
    serialization_default: bool = False
    serialization_if: Optional[Callable[[Any], Any]] = None

    def __call__(
        self,
        deserialization: bool = False,
        serialization: bool = False,
        serialization_default: bool = False,
        serialization_if: Optional[Callable[[Any], Any]] = None,
    ) -> "SkipMetadata":
        return SkipMetadata(
            deserialization, serialization, serialization_default, serialization_if
        )


skip = SkipMetadata(deserialization=True, serialization=True)


@dataclass(frozen=True)
class ValidatorsMetadata(MetadataMixin):
    key = VALIDATORS_METADATA
    validators: Tuple["Validator", ...]


def validators(*validator: Callable) -> ValidatorsMetadata:
    from apischema.validation.validators import Validator

    return ValidatorsMetadata(tuple(map(Validator, validator)))



import re
import warnings
from dataclasses import dataclass, replace
from typing import (
    Any,
    Callable,
    Dict,
    Mapping,
    Optional,
    Pattern,
    Sequence,
    TypeVar,
    Union,
)

from apischema.metadata.keys import SCHEMA_METADATA
from apischema.schemas.annotations import Annotations, ContentEncoding, Deprecated
from apischema.schemas.constraints import Constraints
from apischema.types import AnyType, MetadataMixin, Number, Undefined
from apischema.typing import is_annotated
from apischema.utils import merge_opts, replace_builtins

T = TypeVar("T")
Extra = Union[Mapping[str, Any], Callable[[Dict[str, Any]], None]]


@dataclass(frozen=True)
class Schema(MetadataMixin):
    key = SCHEMA_METADATA
    annotations: Optional[Annotations] = None
    constraints: Optional[Constraints] = None
    extra: Optional[Callable[[Dict[str, Any]], None]] = None
    override: bool = False
    child: Optional["Schema"] = None

    def __call__(self, tp: T) -> T:
        if is_annotated(tp):
            raise TypeError("Cannot register schema on Annotated type")
        _schemas[replace_builtins(tp)] = self
        return tp

    def __set_name__(self, owner, name):
        self.__call__(owner)

    def merge_into(self, base_schema: Dict[str, Any]):
        if self.override:
            base_schema.clear()
        elif self.child is not None:
            self.child.merge_into(base_schema)
        if self.constraints is not None:
            self.constraints.merge_into(base_schema)
        if self.annotations is not None:
            self.annotations.merge_into(base_schema)
        if self.extra is not None:
            self.extra(base_schema)  # type: ignore


def schema(
    *,
    # annotations
    title: Optional[str] = None,
    description: Optional[str] = None,
    default: Any = Undefined,
    examples: Optional[Sequence[Any]] = None,
    deprecated: Optional[Deprecated] = None,
    # number
    min: Optional[Number] = None,
    max: Optional[Number] = None,
    exc_min: Optional[Number] = None,
    exc_max: Optional[Number] = None,
    mult_of: Optional[Number] = None,
    # string
    format: Optional[str] = None,
    media_type: Optional[str] = None,
    encoding: Optional[ContentEncoding] = None,
    min_len: Optional[int] = None,
    max_len: Optional[int] = None,
    pattern: Optional[Union[str, Pattern]] = None,
    # array
    min_items: Optional[int] = None,
    max_items: Optional[int] = None,
    unique: Optional[bool] = None,
    # objects
    min_props: Optional[int] = None,
    max_props: Optional[int] = None,
    # extra
    extra: Optional[Extra] = None,
    override: bool = False,
) -> Schema:
    if default is ...:
        warnings.warn(
            "default=... is deprecated as default value is now"
            " automatically added to the schema",
            DeprecationWarning,
        )
        default = Undefined
    default = None if default is Undefined else (lambda d=default: d)
    if pattern is not None:
        pattern = re.compile(pattern)
    if isinstance(extra, Mapping):
        extra = lambda js, to_update=extra: js.update(to_update)  # type: ignore
    annotations = Annotations(
        title, description, default, examples, format, deprecated, media_type, encoding
    )
    constraints = Constraints(
        min,
        max,
        exc_min,
        exc_max,
        mult_of,
        min_len,
        max_len,
        pattern,
        min_items,
        max_items,
        unique,
        min_props,
        max_props,
    )
    return Schema(annotations, constraints, extra, override)


def default_schema(tp: AnyType) -> Optional[Schema]:
    return None


_schemas: Dict[Any, Schema] = {}


def get_schema(tp: AnyType) -> Optional[Schema]:
    tp = replace_builtins(tp)
    try:
        return _schemas.get(tp)
    except TypeError:
        return None


@merge_opts
def merge_schema(default: Schema, override: Schema) -> Schema:
    if override.override:
        return override
    return replace(override, child=merge_schema(default, override.child))



from enum import Enum
from typing import (
    Any,
    Collection,
    Dict,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
)

from apischema.cache import cache
from apischema.conversions import AnyConversion
from apischema.conversions.conversions import DefaultConversion
from apischema.conversions.visitor import (
    Conv,
    ConversionsVisitor,
    Deserialization,
    DeserializationVisitor,
    Serialization,
    SerializationVisitor,
)
from apischema.objects import ObjectField
from apischema.objects.visitor import (
    DeserializationObjectVisitor,
    ObjectVisitor,
    SerializationObjectVisitor,
)
from apischema.types import AnyType
from apischema.utils import Lazy
from apischema.visitor import Result

RecursionKey = Tuple[AnyType, Optional[AnyConversion]]


class RecursiveChecker(ConversionsVisitor[Conv, Any], ObjectVisitor[Any]):
    def __init__(self, default_conversion: DefaultConversion):
        super().__init__(default_conversion)
        self._cache = recursion_cache(self.__class__)
        self._recursive: Dict[RecursionKey, Set[RecursionKey]] = {}
        self._all_recursive: Set[RecursionKey] = set()
        self._guard: List[RecursionKey] = []
        self._guard_indices: Dict[RecursionKey, int] = {}

    def any(self):
        pass

    def collection(self, cls: Type[Collection], value_type: AnyType):
        return self.visit(value_type)

    def enum(self, cls: Type[Enum]):
        pass

    def literal(self, values: Sequence[Any]):
        pass

    def mapping(self, cls: Type[Mapping], key_type: AnyType, value_type: AnyType):
        self.visit(key_type)
        self.visit(value_type)

    def object(self, tp: AnyType, fields: Sequence[ObjectField]):
        for field in fields:
            self.visit_with_conv(field.type, self._field_conversion(field))

    def primitive(self, cls: Type):
        pass

    def tuple(self, types: Sequence[AnyType]):
        for tp in types:
            self.visit(tp)

    def _visited_union(self, results: Sequence):
        pass

    def unsupported(self, tp: AnyType):
        pass

    def visit(self, tp: AnyType):
        rec_key = (tp, self._conversion)
        if rec_key in self._cache:
            pass
        elif rec_key in self._guard_indices:
            recursive = self._guard[self._guard_indices[rec_key] :]
            self._recursive.setdefault(rec_key, set()).update(recursive)
            self._all_recursive.update(recursive)
        else:
            self._guard_indices[rec_key] = len(self._guard)
            self._guard.append(rec_key)
            try:
                super().visit(tp)
            finally:
                self._guard.pop()
                self._guard_indices.pop(rec_key)
            if rec_key in self._recursive:
                for key in self._recursive[rec_key]:
                    self._cache[key] = True
                assert self._cache[rec_key]
            elif rec_key not in self._all_recursive:
                self._cache[rec_key] = False


class DeserializationRecursiveChecker(
    DeserializationVisitor,
    DeserializationObjectVisitor,
    RecursiveChecker[Deserialization],
):
    pass


class SerializationRecursiveChecker(
    SerializationVisitor, SerializationObjectVisitor, RecursiveChecker[Serialization]
):
    pass


@cache  # use @cache for reset
def recursion_cache(checker_cls: Type[RecursiveChecker]) -> Dict[RecursionKey, bool]:
    return {}


@cache
def is_recursive(
    tp: AnyType,
    conversion: Optional[AnyConversion],
    default_conversion: DefaultConversion,
    checker_cls: Type[RecursiveChecker],
) -> bool:
    cache, rec_key = recursion_cache(checker_cls), (tp, conversion)
    if rec_key not in cache:
        checker_cls(default_conversion).visit_with_conv(tp, conversion)
    return cache[rec_key]


class RecursiveConversionsVisitor(ConversionsVisitor[Conv, Result]):
    def __init__(self, default_conversion: DefaultConversion):
        super().__init__(default_conversion)
        self._cache: Dict[Tuple[AnyType, Optional[AnyConversion]], Result] = {}
        self._first_visit = True

    def _recursive_result(self, lazy: Lazy[Result]) -> Result:
        raise NotImplementedError

    def visit_not_recursive(self, tp: AnyType) -> Result:
        return super().visit(tp)

    def visit(self, tp: AnyType) -> Result:
        if is_recursive(
            tp,
            self._conversion,
            self.default_conversion,
            DeserializationRecursiveChecker  # type: ignore
            if isinstance(self, DeserializationVisitor)
            else SerializationRecursiveChecker,
            # None,
        ):
            cache_key = tp, self._conversion
            if cache_key in self._cache:
                return self._cache[cache_key]
            result = None

            def lazy_result():
                assert result is not None
                return result

            self._cache[cache_key] = self._recursive_result(lazy_result)
            try:
                result = super().visit(tp)
            finally:
                del self._cache[cache_key]
            return result
        elif self._first_visit:
            self._first_visit = False
            return super().visit(tp)
        else:
            return self.visit_not_recursive(tp)



from contextlib import contextmanager, suppress
from dataclasses import replace
from functools import lru_cache
from types import new_class
from typing import (
    Any,
    Collection,
    Generic,
    Iterable,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from apischema.conversions import LazyConversion
from apischema.conversions.conversions import (
    AnyConversion,
    DefaultConversion,
    ResolvedConversion,
    ResolvedConversions,
    handle_identity_conversion,
    is_identity,
    resolve_any_conversion,
)
from apischema.conversions.utils import is_convertible
from apischema.metadata.implem import ConversionMetadata
from apischema.metadata.keys import CONVERSION_METADATA
from apischema.type_names import type_name
from apischema.types import AnyType
from apischema.typing import get_args, is_type_var
from apischema.utils import (
    context_setter,
    get_args2,
    get_origin_or_type,
    has_type_vars,
    is_subclass,
    substitute_type_vars,
    subtyping_substitution,
)
from apischema.visitor import Result, Unsupported, Visitor

Deserialization = ResolvedConversions
Serialization = ResolvedConversion
Conv = TypeVar("Conv")


class ConversionsVisitor(Visitor[Result], Generic[Conv, Result]):
    def __init__(self, default_conversion: DefaultConversion):
        self.default_conversion = default_conversion
        self._conversion: Optional[AnyConversion] = None

    def _has_conversion(
        self, tp: AnyType, conversion: Optional[AnyConversion]
    ) -> Tuple[bool, Optional[Conv]]:
        raise NotImplementedError

    def _annotated_conversion(
        self, annotation: ConversionMetadata
    ) -> Optional[AnyConversion]:
        raise NotImplementedError

    def annotated(self, tp: AnyType, annotations: Sequence[Any]) -> Result:
        for annotation in reversed(annotations):
            if isinstance(annotation, Mapping) and CONVERSION_METADATA in annotation:
                with self._replace_conversion(
                    self._annotated_conversion(annotation[CONVERSION_METADATA])
                ):
                    return super().annotated(tp, annotations)
        return super().annotated(tp, annotations)

    def _union_results(self, alternatives: Iterable[AnyType]) -> Sequence[Result]:
        results = []
        for alt in alternatives:
            with suppress(Unsupported):
                results.append(self.visit(alt))
        if not results:
            raise Unsupported(Union[tuple(alternatives)])
        return results

    def _visited_union(self, results: Sequence[Result]) -> Result:
        raise NotImplementedError

    def union(self, alternatives: Sequence[AnyType]) -> Result:
        return self._visited_union(self._union_results(alternatives))

    @contextmanager
    def _replace_conversion(self, conversion: Optional[AnyConversion]):
        with context_setter(self):
            self._conversion = resolve_any_conversion(conversion) or None
            yield

    def visit_with_conv(
        self, tp: AnyType, conversion: Optional[AnyConversion]
    ) -> Result:
        with self._replace_conversion(conversion):
            return self.visit(tp)

    def _visit_conversion(
        self,
        tp: AnyType,
        conversion: Conv,
        dynamic: bool,
        next_conversion: Optional[AnyConversion],
    ) -> Result:
        raise NotImplementedError

    def visit_conversion(
        self,
        tp: AnyType,
        conversion: Optional[Conv],
        dynamic: bool,
        next_conversion: Optional[AnyConversion] = None,
    ) -> Result:
        if conversion is not None:
            return self._visit_conversion(tp, conversion, dynamic, next_conversion)
        else:
            with self._replace_conversion(next_conversion):
                return super().visit(tp)

    def visit(self, tp: AnyType) -> Result:
        if not is_convertible(tp):
            return self.visit_conversion(tp, None, False, self._conversion)
        dynamic, conversion = self._has_conversion(tp, self._conversion)
        if not dynamic:
            _, conversion = self._has_conversion(
                tp, self.default_conversion(get_origin_or_type(tp))  # type: ignore
            )
        next_conversion = None
        if not dynamic and is_subclass(tp, Collection):
            next_conversion = self._conversion
        return self.visit_conversion(tp, conversion, dynamic, next_conversion)


def sub_conversion(
    conversion: ResolvedConversion, next_conversion: Optional[AnyConversion]
) -> Optional[AnyConversion]:
    return (
        LazyConversion(lambda: conversion.sub_conversion),
        LazyConversion(lambda: next_conversion),
    )


@lru_cache(maxsize=0)
def self_deserialization_wrapper(cls: Type) -> Type:
    wrapper = new_class(
        f"{cls.__name__}SelfDeserializer",
        (cls[cls.__parameters__] if has_type_vars(cls) else cls,),
        exec_body=lambda ns: ns.update(
            {"__new__": lambda _, *args, **kwargs: cls(*args, **kwargs)}
        ),
    )
    return type_name(None)(wrapper)


class DeserializationVisitor(ConversionsVisitor[Deserialization, Result]):
    @staticmethod
    def _has_conversion(
        tp: AnyType, conversion: Optional[AnyConversion]
    ) -> Tuple[bool, Optional[Deserialization]]:
        identity_conv, result = False, []
        for conv in resolve_any_conversion(conversion):
            conv = handle_identity_conversion(conv, tp)
            if is_subclass(conv.target, tp):
                if is_identity(conv):
                    if identity_conv:
                        continue
                    identity_conv = True
                    wrapper: AnyType = self_deserialization_wrapper(
                        get_origin_or_type(tp)
                    )
                    if get_args(tp):
                        wrapper = wrapper[get_args(tp)]
                    conv = ResolvedConversion(replace(conv, source=wrapper))
                if is_type_var(conv.source) or any(
                    map(is_type_var, get_args2(conv.source))
                ):
                    _, substitution = subtyping_substitution(tp, conv.target)
                    conv = replace(
                        conv, source=substitute_type_vars(conv.source, substitution)
                    )
                result.append(ResolvedConversion(replace(conv, target=tp)))
        if identity_conv and len(result) == 1:
            return True, None
        else:
            return bool(result), tuple(result) or None

    def _annotated_conversion(
        self, annotation: ConversionMetadata
    ) -> Optional[AnyConversion]:
        return annotation.deserialization

    def _visit_conversion(
        self,
        tp: AnyType,
        conversion: Deserialization,
        dynamic: bool,
        next_conversion: Optional[AnyConversion],
    ) -> Result:
        results = [
            self.visit_with_conv(conv.source, sub_conversion(conv, next_conversion))
            for conv in conversion
        ]
        return self._visited_union(results)


class SerializationVisitor(ConversionsVisitor[Serialization, Result]):
    @staticmethod
    def _has_conversion(
        tp: AnyType, conversion: Optional[AnyConversion]
    ) -> Tuple[bool, Optional[Serialization]]:
        for conv in resolve_any_conversion(conversion):
            conv = handle_identity_conversion(conv, tp)
            if is_subclass(tp, conv.source):
                if is_identity(conv):
                    return True, None
                if is_type_var(conv.target) or any(
                    map(is_type_var, get_args2(conv.target))
                ):
                    substitution, _ = subtyping_substitution(conv.source, tp)
                    conv = replace(
                        conv, target=substitute_type_vars(conv.target, substitution)
                    )
                return True, ResolvedConversion(replace(conv, source=tp))
        else:
            return False, None

    def _annotated_conversion(
        self, annotation: ConversionMetadata
    ) -> Optional[AnyConversion]:
        return annotation.serialization

    def _visit_conversion(
        self,
        tp: AnyType,
        conversion: Serialization,
        dynamic: bool,
        next_conversion: Optional[AnyConversion],
    ) -> Result:
        return self.visit_with_conv(
            conversion.target, sub_conversion(conversion, next_conversion)
        )



from typing import Callable, MutableMapping, TypeVar, overload

from apischema.cache import CacheAwareDict
from apischema.types import Metadata, MetadataImplem

Aliaser = Callable[[str], str]
Cls = TypeVar("Cls", bound=type)

_class_aliasers: MutableMapping[type, Aliaser] = CacheAwareDict({})

get_class_aliaser = _class_aliasers.get


@overload
def alias(alias_: str, *, override: bool = True) -> Metadata:
    ...


@overload
def alias(override: bool) -> Metadata:
    ...


@overload
def alias(aliaser: Aliaser) -> Callable[[Cls], Cls]:
    ...


def alias(arg=None, *, override: bool = True):  # type: ignore
    """Field alias or class aliaser

    :param alias_: alias of the field
    :param override: alias can be overridden by a class aliaser
    :param aliaser: compute alias for each (overridable) field of the class decorated
    """
    from apischema.metadata.keys import ALIAS_METADATA, ALIAS_NO_OVERRIDE_METADATA

    if callable(arg):

        def aliaser(cls: Cls) -> Cls:
            _class_aliasers[cls] = arg
            return cls

        return aliaser
    else:
        metadata = MetadataImplem()
        if arg is not None:
            metadata[ALIAS_METADATA] = arg
        if not override:
            metadata[ALIAS_NO_OVERRIDE_METADATA] = True
        if not metadata:
            raise NotImplementedError
        return metadata



import inspect
from typing import (
    Any,
    Callable,
    Mapping,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
    cast,
    overload,
)

from apischema.cache import cache
from apischema.metadata import properties
from apischema.objects.fields import ObjectField
from apischema.objects.visitor import ObjectVisitor
from apischema.types import AnyType, OrderedDict
from apischema.typing import _GenericAlias, get_type_hints
from apischema.utils import empty_dict
from apischema.visitor import Unsupported


@cache
def object_fields(
    tp: AnyType,
    deserialization: bool = False,
    serialization: bool = False,
    default: Optional[
        Callable[[type], Optional[Sequence[ObjectField]]]
    ] = ObjectVisitor._default_fields,
) -> Mapping[str, ObjectField]:
    class GetFields(ObjectVisitor[Sequence[ObjectField]]):
        def _skip_field(self, field: ObjectField) -> bool:
            return (field.skip.deserialization and serialization) or (
                field.skip.serialization and deserialization
            )

        @staticmethod
        def _default_fields(cls: type) -> Optional[Sequence[ObjectField]]:
            return None if default is None else default(cls)

        def object(
            self, cls: Type, fields: Sequence[ObjectField]
        ) -> Sequence[ObjectField]:
            return fields

    try:
        return OrderedDict((f.name, f) for f in GetFields().visit(tp))
    except (Unsupported, NotImplementedError):
        raise TypeError(f"{tp} doesn't have fields")


def object_fields2(obj: Any) -> Mapping[str, ObjectField]:
    return object_fields(
        obj if isinstance(obj, (type, _GenericAlias)) else obj.__class__
    )


T = TypeVar("T")


class FieldGetter:
    def __init__(self, obj: Any):
        self.fields = object_fields2(obj)

    def __getattribute__(self, name: str) -> ObjectField:
        try:
            return object.__getattribute__(self, "fields")[name]
        except KeyError:
            raise AttributeError(name)


@overload
def get_field(obj: Type[T]) -> T:
    ...


@overload
def get_field(obj: T) -> T:
    ...


# Overload because of Mypy issue
# https://github.com/python/mypy/issues/9003#issuecomment-667418520
def get_field(obj: Union[Type[T], T]) -> T:
    return cast(T, FieldGetter(obj))


class AliasedStr(str):
    pass


class AliasGetter:
    def __init__(self, obj: Any):
        self.fields = object_fields2(obj)

    def __getattribute__(self, name: str) -> str:
        try:
            return AliasedStr(object.__getattribute__(self, "fields")[name].alias)
        except KeyError:
            raise AttributeError(name)


@overload
def get_alias(obj: Type[T]) -> T:
    ...


@overload
def get_alias(obj: T) -> T:
    ...


def get_alias(obj: Union[Type[T], T]) -> T:
    return cast(T, AliasGetter(obj))


def parameters_as_fields(
    func: Callable, parameters_metadata: Mapping[str, Mapping] = None
) -> Sequence[ObjectField]:
    parameters_metadata = parameters_metadata or {}
    types = get_type_hints(func, include_extras=True)
    fields = []
    for param_name, param in inspect.signature(func).parameters.items():
        if param.kind is inspect.Parameter.POSITIONAL_ONLY:
            raise TypeError("Positional only parameters are not supported")
        param_type = types.get(param_name, Any)
        if param.kind in {
            inspect.Parameter.POSITIONAL_OR_KEYWORD,
            inspect.Parameter.KEYWORD_ONLY,
        }:
            field = ObjectField(
                param_name,
                param_type,
                param.default is inspect.Parameter.empty,
                parameters_metadata.get(param_name, empty_dict),
                default=param.default,
            )
            fields.append(field)
        elif param.kind == inspect.Parameter.VAR_KEYWORD:
            field = ObjectField(
                param_name,
                Mapping[str, param_type],  # type: ignore
                False,
                properties | parameters_metadata.get(param_name, empty_dict),
                default_factory=dict,
            )
            fields.append(field)
    return fields



from contextlib import suppress
from typing import (
    Any,
    Collection,
    Iterable,
    Iterator,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    Union,
)

from apischema.conversions.conversions import AnyConversion, DefaultConversion
from apischema.conversions.visitor import (
    Conv,
    ConversionsVisitor,
    DeserializationVisitor,
    SerializationVisitor,
)
from apischema.types import AnyType
from apischema.utils import is_hashable
from apischema.visitor import Unsupported

try:
    from apischema.typing import Annotated, is_union
except ImportError:
    Annotated = ...  # type: ignore


def merge_results(
    results: Iterable[Sequence[AnyType]], origin: AnyType
) -> Sequence[AnyType]:
    def rec(index=0) -> Iterator[Sequence[AnyType]]:
        if index < len(result_list):
            for next_ in rec(index + 1):
                for res in result_list[index]:
                    yield (res, *next_)
        else:
            yield ()

    result_list = list(results)
    return [(Union if is_union(origin) else origin)[tuple(r)] for r in rec()]


class ConversionsResolver(ConversionsVisitor[Conv, Sequence[AnyType]]):
    def __init__(self, default_conversion: DefaultConversion):
        super().__init__(default_conversion)
        self._skip_conversion = True
        self._rec_guard: Set[Tuple[AnyType, Conv]] = set()

    def annotated(self, tp: AnyType, annotations: Sequence[Any]) -> Sequence[AnyType]:
        return [
            Annotated[(res, *annotations)] for res in super().annotated(tp, annotations)
        ]

    def collection(
        self, cls: Type[Collection], value_type: AnyType
    ) -> Sequence[AnyType]:
        return merge_results([self.visit(value_type)], Collection)

    def mapping(
        self, cls: Type[Mapping], key_type: AnyType, value_type: AnyType
    ) -> Sequence[AnyType]:
        return merge_results([self.visit(key_type), self.visit(value_type)], Mapping)

    def new_type(self, tp: AnyType, super_type: AnyType) -> Sequence[AnyType]:
        raise NotImplementedError

    def tuple(self, types: Sequence[AnyType]) -> Sequence[AnyType]:
        return merge_results(map(self.visit, types), Tuple)

    def _visited_union(self, results: Sequence[Sequence[AnyType]]) -> Sequence[AnyType]:
        return merge_results(results, Union)

    def visit_conversion(
        self,
        tp: AnyType,
        conversion: Any,
        dynamic: bool,
        next_conversion: Optional[AnyConversion] = None,
    ) -> Sequence[AnyType]:
        if conversion is not None and self._skip_conversion:
            return [] if dynamic else [tp]
        self._skip_conversion = False
        results: Sequence[AnyType] = []
        if not is_hashable(tp):
            with suppress(NotImplementedError, Unsupported):
                results = super().visit_conversion(
                    tp, conversion, dynamic, next_conversion
                )
        elif (tp, conversion) not in self._rec_guard:
            self._rec_guard.add((tp, conversion))
            with suppress(NotImplementedError, Unsupported):
                results = super().visit_conversion(
                    tp, conversion, dynamic, next_conversion
                )
            self._rec_guard.remove((tp, conversion))
        if not dynamic and (conversion is not None or not results):
            results = [tp, *results]
        return results


class WithConversionsResolver:
    def resolve_conversion(self, tp: AnyType) -> Sequence[AnyType]:
        raise NotImplementedError

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        Resolver: Type[ConversionsResolver]
        if issubclass(cls, DeserializationVisitor):

            class Resolver(ConversionsResolver, DeserializationVisitor):
                pass

        elif issubclass(cls, SerializationVisitor):

            class Resolver(ConversionsResolver, SerializationVisitor):
                pass

        else:
            return

        def resolve_conversion(
            self: ConversionsVisitor, tp: AnyType
        ) -> Sequence[AnyType]:
            return Resolver(self.default_conversion).visit_with_conv(
                tp, self._conversion
            )

        assert issubclass(cls, WithConversionsResolver)
        cls.resolve_conversion = resolve_conversion  # type: ignore



import warnings
from dataclasses import dataclass
from types import new_class
from typing import Callable, Optional, TYPE_CHECKING, Tuple, Type, Union

from apischema.conversions import Conversion
from apischema.conversions.conversions import ResolvedConversion
from apischema.dataclasses import replace
from apischema.utils import PREFIX, identity

if TYPE_CHECKING:
    from apischema.deserialization.coercion import Coerce

Model = Union[Type, Callable[[], Type]]


def check_model(origin: Type, model: Type):
    if not isinstance(model, type):
        raise TypeError("Dataclass model must be a dataclass")
    if getattr(origin, "__parameters__", ()) != getattr(model, "__parameters__", ()):
        raise TypeError("Dataclass model must have the same generic parameters")


MODEL_ORIGIN_ATTR = f"{PREFIX}model_origin"

DATACLASS_ATTR = "_dataclass"


@dataclass(frozen=True)
class DataclassModel:
    origin: Type
    model: Model
    fields_only: bool

    @property
    def dataclass(self) -> Type:
        if not hasattr(self, "_dataclass"):
            origin = self.origin
            if isinstance(self.model, type):
                assert check_model(origin, self.model) is None
                model = self.model
            else:
                model = self.model()
                check_model(origin, model)
            namespace = {"__new__": lambda _, *args, **kwargs: origin(*args, **kwargs)}
            if not self.fields_only:
                namespace[MODEL_ORIGIN_ATTR] = origin
            cls = new_class(
                model.__name__, (model,), exec_body=lambda ns: ns.update(namespace)
            )
            object.__setattr__(self, "_dataclass", cls)
        return getattr(self, "_dataclass")


def dataclass_model(
    origin: Type,
    model: Model,
    *,
    fields_only: bool = False,
    additional_properties: Optional[bool] = None,
    coercion: Optional["Coerce"] = None,
    fall_back_on_default: Optional[bool] = None,
    exclude_unset: Optional[bool] = None,
) -> Tuple[Conversion, Conversion]:
    warnings.warn(
        "dataclass_model is deprecated, use set_object_fields instead",
        DeprecationWarning,
    )
    if isinstance(model, type):
        check_model(origin, model)

    model_type = DataclassModel(origin, model, fields_only)
    return Conversion(identity, source=model_type, target=origin), Conversion(
        identity, source=origin, target=model_type
    )


def has_model_origin(cls: Type) -> bool:
    return hasattr(cls, MODEL_ORIGIN_ATTR)


def get_model_origin(cls: Type) -> Type:
    return getattr(cls, MODEL_ORIGIN_ATTR)


def handle_dataclass_model(conversion: ResolvedConversion) -> ResolvedConversion:
    conv: Conversion = conversion
    if isinstance(conv.source, DataclassModel):
        conv = replace(conv, source=conv.source.dataclass)
    if isinstance(conv.target, DataclassModel):
        conv = replace(conv, target=conv.target.dataclass)
    return ResolvedConversion(conv)



__all__ = [
    "Discard",
    "LocalizedError",
    "ValidationError",
    "ValidatorResult",
    "get_validators",
    "validate",
    "validator",
]

from .errors import LocalizedError, ValidationError, ValidatorResult
from .validators import Discard, get_validators, validate, validator



from apischema.utils import PREFIX

ALIAS_METADATA = f"{PREFIX}alias"
ALIAS_NO_OVERRIDE_METADATA = f"{PREFIX}alias_no_override"
CONVERSION_METADATA = f"{PREFIX}conversion"
DEFAULT_AS_SET_METADATA = f"{PREFIX}default_as_set"
FALL_BACK_ON_DEFAULT_METADATA = f"{PREFIX}fall_back_on_default"
FLATTEN_METADATA = f"{PREFIX}flattened"
INIT_VAR_METADATA = f"{PREFIX}init_var"
NONE_AS_UNDEFINED_METADATA = f"{PREFIX}none_as_undefined"
ORDERING_METADATA = f"{PREFIX}ordering"
POST_INIT_METADATA = f"{PREFIX}post_init"
PROPERTIES_METADATA = f"{PREFIX}properties"
REQUIRED_METADATA = f"{PREFIX}required"
SCHEMA_METADATA = f"{PREFIX}schema"
SKIP_METADATA = f"{PREFIX}skip"
VALIDATORS_METADATA = f"{PREFIX}validators"



from typing import Collection, Set, Type, TypeVar

_interfaces: Set[Type] = set()

Cls = TypeVar("Cls", bound=Type)


def interface(cls: Cls) -> Cls:
    _interfaces.add(cls)
    return cls


def is_interface(cls: Type) -> bool:
    return cls in _interfaces


def get_interfaces(cls: Type) -> Collection[Type]:
    return list(filter(is_interface, cls.__mro__[1:]))



import operator as op
from collections import defaultdict
from dataclasses import dataclass, field, fields
from math import gcd
from typing import (
    Any,
    Callable,
    Collection,
    Dict,
    Mapping,
    Optional,
    Pattern,
    Tuple,
    TypeVar,
)

from apischema.types import Number
from apischema.utils import merge_opts, to_hashable

T = TypeVar("T")
U = TypeVar("U")

COMPARISON_MERGE_AND_ERRORS: Dict[Callable, Tuple[Callable, str]] = {
    op.lt: (max, "less than %s"),
    op.le: (max, "less than or equal to %s"),
    op.gt: (min, "greater than %s"),
    op.ge: (min, "greater than or equal to %s"),
}
PREFIX_DICT: Mapping[type, str] = {
    str: "string length",
    list: "item count",
    dict: "property count",
}
Check = Callable[[Any, Any], Any]
CONSTRAINT_METADATA_KEY = "constraint"


@dataclass
class ConstraintMetadata:
    alias: str
    cls: type
    check: Check
    error: Callable[[Any], str]
    merge: Callable[[T, T], T]

    @property
    def field(self) -> Any:
        return field(default=None, metadata={CONSTRAINT_METADATA_KEY: self})


def comparison(alias: str, cls: type, check: Check) -> Any:
    merge, error = COMPARISON_MERGE_AND_ERRORS[check]
    prefix = PREFIX_DICT.get(cls)  # type: ignore
    if prefix:
        error = prefix + " " + error.replace("less", "lower")
    if cls in (str, list, dict):
        wrapped = check

        def check(data: Any, value: Any) -> bool:
            return wrapped(len(data), value)

    return ConstraintMetadata(alias, cls, check, lambda v: error % v, merge).field


def merge_mult_of(m1: Number, m2: Number) -> Number:
    if not isinstance(m1, int) and not isinstance(m2, int):
        raise TypeError("multipleOf merging is only supported with integers")
    return m1 * m2 / gcd(m1, m2)  # type: ignore


def not_match_pattern(data: str, pattern: Pattern) -> bool:
    return not pattern.match(data)


def merge_pattern(p1: Pattern, p2: Pattern) -> Pattern:
    raise TypeError("Cannot merge patterns")


def not_unique(data: list, unique: bool) -> bool:
    return (op.ne if unique else op.eq)(len(set(map(to_hashable, data))), len(data))


@dataclass(frozen=True)
class Constraints:
    # number
    min: Optional[Number] = comparison("minimum", float, op.lt)
    max: Optional[Number] = comparison("maximum", float, op.gt)
    exc_min: Optional[Number] = comparison("exclusiveMinimum", float, op.le)
    exc_max: Optional[Number] = comparison("exclusiveMaximum", float, op.ge)
    mult_of: Optional[Number] = ConstraintMetadata(
        "multipleOf", float, op.mod, lambda n: f"not a multiple of {n}", merge_mult_of  # type: ignore
    ).field
    # string
    min_len: Optional[int] = comparison("minLength", str, op.lt)
    max_len: Optional[int] = comparison("maxLength", str, op.gt)
    pattern: Optional[Pattern] = ConstraintMetadata(
        "pattern",
        str,
        not_match_pattern,
        lambda p: f"not matching '{p.pattern}'",
        merge_pattern,  # type: ignore
    ).field
    # array
    min_items: Optional[int] = comparison("minItems", list, op.lt)
    max_items: Optional[int] = comparison("maxItems", list, op.gt)
    unique: Optional[bool] = ConstraintMetadata(
        "uniqueItems", list, not_unique, lambda _: "duplicate items", op.or_
    ).field
    # object
    min_props: Optional[int] = comparison("minProperties", dict, op.lt)
    max_props: Optional[int] = comparison("maxProperties", dict, op.gt)

    @property
    def attr_and_metata(
        self,
    ) -> Collection[Tuple[str, Optional[Any], ConstraintMetadata]]:
        return [
            (f.name, getattr(self, f.name), f.metadata[CONSTRAINT_METADATA_KEY])
            for f in fields(self)
            if CONSTRAINT_METADATA_KEY in f.metadata
        ]

    @property
    def checks_by_type(self) -> Mapping[type, Collection[Tuple[Check, Any, str]]]:
        result = defaultdict(list)
        for _, attr, metadata in self.attr_and_metata:
            if attr is None:
                continue
            error = f"{metadata.error(attr)} ({metadata.alias})"
            result[metadata.cls].append((metadata.check, attr, error))
        result[int] = result[float]
        return result

    def merge_into(self, base_schema: Dict[str, Any]):
        for name, attr, metadata in self.attr_and_metata:
            if attr is not None:
                alias = metadata.alias
                if alias in base_schema:
                    base_schema[alias] = metadata.merge(attr, base_schema[alias])  # type: ignore
                else:
                    base_schema[alias] = attr


@merge_opts
def merge_constraints(c1: Constraints, c2: Constraints) -> Constraints:
    constraints: Dict[str, Any] = {}
    for name, attr1, metadata in c1.attr_and_metata:
        attr2 = getattr(c2, name)
        if attr1 is None:
            constraints[name] = attr2
        elif attr2 is None:
            constraints[name] = attr1
        else:
            constraints[name] = metadata.merge(attr1, attr2)  # type: ignore
    return Constraints(**constraints)  # type: ignore



import warnings
from inspect import Parameter
from typing import Callable, Optional, Sequence

from apischema import cache
from apischema.aliases import Aliaser
from apischema.conversions.conversions import DefaultConversion
from apischema.conversions.converters import (
    default_deserialization,
    default_serialization,
)
from apischema.deserialization.coercion import Coercer, coerce as coerce_
from apischema.json_schema import JsonSchemaVersion
from apischema.objects import ObjectField
from apischema.objects.fields import default_object_fields as default_object_fields_
from apischema.schemas import Schema
from apischema.serialization import PassThroughOptions
from apischema.type_names import TypeName, default_type_name as default_type_name_
from apischema.types import AnyType
from apischema.utils import to_camel_case


class ResetCache(type):
    def __setattr__(self, name, value):
        super().__setattr__(name, value)
        cache.reset()


class MetaSettings(ResetCache):
    @property
    def camel_case(self) -> bool:
        raise NotImplementedError

    @camel_case.setter
    def camel_case(self, value: bool):
        settings.aliaser = to_camel_case if value else lambda s: s

    def __setattr__(self, name, value):
        if name == "default_schema" and not isinstance(value, ResetCache):
            warnings.warn(
                "settings.default_schema is deprecated,"
                " use settings.base_schema.type instead",
                DeprecationWarning,
            )
            assert self is settings
            self.base_schema.type = value  # type: ignore
        else:
            super().__setattr__(name, value)


class settings(metaclass=MetaSettings):
    additional_properties: bool = False
    aliaser: Aliaser = lambda s: s
    default_object_fields: Callable[
        [type], Optional[Sequence[ObjectField]]
    ] = default_object_fields_
    default_schema: Callable[[AnyType], Optional[Schema]] = lambda *_: None
    default_type_name: Callable[[AnyType], Optional[TypeName]] = default_type_name_
    json_schema_version: JsonSchemaVersion = JsonSchemaVersion.DRAFT_2020_12

    class base_schema:
        field: Callable[[AnyType, str, str], Optional[Schema]] = lambda *_: None
        method: Callable[[AnyType, Callable, str], Optional[Schema]] = lambda *_: None
        parameter: Callable[
            [Callable, Parameter, str], Optional[Schema]
        ] = lambda *_: None
        type: Callable[[AnyType], Optional[Schema]] = lambda *_: None

    class deserialization(metaclass=ResetCache):
        coerce: bool = False
        coercer: Coercer = coerce_
        default_conversion: DefaultConversion = default_deserialization
        fall_back_on_default: bool = False

    class serialization(metaclass=ResetCache):
        check_type: bool = False
        fall_back_on_any: bool = False
        default_conversion: DefaultConversion = default_serialization
        exclude_defaults: bool = False
        exclude_none: bool = False
        exclude_unset: bool = True
        pass_through: PassThroughOptions = PassThroughOptions()



import inspect
from functools import wraps
from inspect import signature
from types import FunctionType
from typing import Callable, Generic, Optional, Type, Union, cast

from apischema.typing import get_type_hints
from apischema.utils import PREFIX, T, get_origin_or_type2

MethodOrProperty = Union[Callable, property]


def _method_location(method: MethodOrProperty) -> Optional[Type]:
    if isinstance(method, property):
        assert method.fget is not None
        method = method.fget
    while hasattr(method, "__wrapped__"):
        method = method.__wrapped__  # type: ignore
    assert isinstance(method, FunctionType)
    global_name, *class_path = method.__qualname__.split(".")[:-1]
    if global_name not in method.__globals__:
        return None
    location = method.__globals__[global_name]
    for attr in class_path:
        if hasattr(location, attr):
            location = getattr(location, attr)
        else:
            break
    return location


def is_method(method: MethodOrProperty) -> bool:
    """Return if the function is method/property declared in a class"""
    return (
        isinstance(method, property)
        and method.fget is not None
        and is_method(method.fget)
    ) or (
        isinstance(method, FunctionType)
        and method.__name__ != method.__qualname__
        and isinstance(_method_location(method), (type, type(None)))
        and next(iter(inspect.signature(method).parameters), None) == "self"
    )


def method_class(method: MethodOrProperty) -> Optional[Type]:
    cls = _method_location(method)
    return cls if isinstance(cls, type) else None


METHOD_WRAPPER_ATTR = f"{PREFIX}method_wrapper"


def method_wrapper(method: MethodOrProperty, name: str = None) -> Callable:
    if isinstance(method, property):
        assert method.fget is not None
        name = name or method.fget.__name__

        @wraps(method.fget)
        def wrapper(self):
            return getattr(self, name)

    else:
        if hasattr(method, METHOD_WRAPPER_ATTR):
            return method
        name = name or method.__name__

        if list(signature(method).parameters) == ["self"]:

            @wraps(method)
            def wrapper(self):
                return getattr(self, name)()

        else:

            @wraps(method)
            def wrapper(self, *args, **kwargs):
                return getattr(self, name)(*args, **kwargs)

    setattr(wrapper, METHOD_WRAPPER_ATTR, True)
    return wrapper


class MethodWrapper(Generic[T]):
    def __init__(self, method: T):
        self._method = method

    def getter(self, func):
        self._method = self._method.getter(func)
        return self

    def setter(self, func):
        self._method = self._method.setter(func)
        return self

    def deleter(self, func):
        self._method = self._method.deleter(func)
        return self

    def __set_name__(self, owner, name):
        setattr(owner, name, self._method)

    def __call__(self, *args, **kwargs):
        raise RuntimeError("Method __set_name__ has not been called")


def method_registerer(
    arg: Optional[Callable],
    owner: Optional[Type],
    register: Callable[[Callable, Type, str], None],
):
    def decorator(method: MethodOrProperty):
        if owner is None and is_method(method) and method_class(method) is None:

            class Descriptor(MethodWrapper[MethodOrProperty]):
                def __set_name__(self, owner, name):
                    super().__set_name__(owner, name)
                    register(method_wrapper(method), owner, name)

            return Descriptor(method)
        else:
            owner2 = owner
            if is_method(method):
                if owner2 is None:
                    owner2 = method_class(method)
                method = method_wrapper(method)
            if owner2 is None:
                try:
                    hints = get_type_hints(method)
                    owner2 = get_origin_or_type2(hints[next(iter(hints))])
                except (KeyError, StopIteration):
                    raise TypeError("First parameter of method must be typed") from None
            assert not isinstance(method, property)
            register(cast(Callable, method), owner2, method.__name__)
            return method

    return decorator if arg is None else decorator(arg)



from dataclasses import dataclass
from functools import lru_cache
from typing import (
    Any,
    Callable,
    Collection,
    Dict,
    Generic,
    List,
    NewType,
    Optional,
    TYPE_CHECKING,
    Tuple,
    TypeVar,
    Union,
)

from apischema.conversions.utils import Converter, converter_types
from apischema.dataclasses import replace
from apischema.methods import is_method, method_class, method_wrapper
from apischema.types import AnyType
from apischema.typing import is_type_var
from apischema.utils import deprecate_kwargs, identity

if TYPE_CHECKING:
    pass

ConvOrProp = TypeVar("ConvOrProp", Converter, property)


@dataclass(frozen=True)
class Conversion(Generic[ConvOrProp]):
    converter: ConvOrProp
    source: AnyType = None
    target: AnyType = None
    sub_conversion: Optional["AnyConversion"] = None
    inherited: Optional[bool] = None


deprecate_kwargs({"sub_conversions": "sub_conversion"})(Conversion)


@dataclass(frozen=True)
class LazyConversion:
    get: Callable[[], Optional["AnyConversion"]]

    def __post_init__(self):
        object.__setattr__(self, "get", lru_cache(1)(self.get))

    @property
    def inherited(self) -> Optional[bool]:
        conversion = self.get()  # type: ignore
        return isinstance(conversion, Conversion) and conversion.inherited


ConvOrFunc = Union[Conversion, Converter, property, LazyConversion]
AnyConversion = Union[ConvOrFunc, Tuple[ConvOrFunc, ...]]
DefaultConversion = Callable[[AnyType], Optional[AnyConversion]]


ResolvedConversion = NewType("ResolvedConversion", Conversion[Converter])
ResolvedConversions = Tuple[ResolvedConversion, ...]  # Tuple in order to be hashable


def resolve_conversion(
    conversion: Union[Converter, property, Conversion], namespace: Dict[str, Any] = None
) -> ResolvedConversion:
    if not isinstance(conversion, Conversion):
        conversion = Conversion(conversion)
    if is_method(conversion.converter):
        if conversion.source is None:
            conversion = replace(conversion, source=method_class(conversion.converter))
        conversion = replace(conversion, converter=method_wrapper(conversion.converter))
    assert not isinstance(conversion.converter, property)
    source, target = converter_types(
        conversion.converter, conversion.source, conversion.target, namespace
    )
    return ResolvedConversion(replace(conversion, source=source, target=target))


def resolve_any_conversion(conversion: Optional[AnyConversion]) -> ResolvedConversions:
    if not conversion:
        return ()
    result: List[ResolvedConversion] = []
    for conv in conversion if isinstance(conversion, Collection) else [conversion]:
        if isinstance(conv, LazyConversion):
            result.extend(resolve_any_conversion(conv.get()))  # type: ignore
        else:
            result.append(resolve_conversion(conv))
    return tuple(result)


def handle_identity_conversion(
    conversion: ResolvedConversion, tp: AnyType
) -> ResolvedConversion:
    if (
        is_identity(conversion)
        and conversion.source == conversion.target
        and is_type_var(conversion.source)
    ):
        return ResolvedConversion(replace(conversion, source=tp, target=tp))
    else:
        return conversion


def is_identity(conversion: ResolvedConversion) -> bool:
    return (
        conversion.converter == identity
        and conversion.source == conversion.target
        and conversion.sub_conversion is None
    )



from collections import defaultdict
from enum import Enum
from typing import (
    Any,
    Collection,
    Dict,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
)

from apischema.conversions.conversions import AnyConversion, DefaultConversion
from apischema.conversions.visitor import (
    ConversionsVisitor,
    DeserializationVisitor,
    SerializationVisitor,
)
from apischema.json_schema.conversions_resolver import WithConversionsResolver
from apischema.objects import ObjectField
from apischema.objects.visitor import (
    DeserializationObjectVisitor,
    ObjectVisitor,
    SerializationObjectVisitor,
)
from apischema.type_names import TypeNameFactory, get_type_name
from apischema.types import AnyType
from apischema.utils import is_hashable, replace_builtins
from apischema.visitor import Unsupported

try:
    from apischema.typing import Annotated
except ImportError:
    Annotated = ...  # type: ignore

Refs = Dict[str, Tuple[AnyType, int]]


class Recursive(Exception):
    pass


T = TypeVar("T")


class RefsExtractor(ConversionsVisitor, ObjectVisitor, WithConversionsResolver):
    def __init__(self, default_conversion: DefaultConversion, refs: Refs):
        super().__init__(default_conversion)
        self.refs = refs
        self._rec_guard: Dict[
            Tuple[AnyType, Optional[AnyConversion]], int
        ] = defaultdict(lambda: 0)

    def _incr_ref(self, ref: Optional[str], tp: AnyType) -> bool:
        if ref is None:
            return False
        else:
            ref_cls, count = self.refs.get(ref, (tp, 0))
            if replace_builtins(ref_cls) != replace_builtins(tp):
                raise ValueError(
                    f"Types {tp} and {self.refs[ref][0]} share same reference '{ref}'"
                )
            self.refs[ref] = (ref_cls, count + 1)
            return count > 0

    def annotated(self, tp: AnyType, annotations: Sequence[Any]):
        for i, annotation in enumerate(reversed(annotations)):
            if isinstance(annotation, TypeNameFactory):
                ref = annotation.to_type_name(tp).json_schema
                if not isinstance(ref, str):
                    continue
                ref_annotations = annotations[: len(annotations) - i]
                annotated = Annotated[(tp, *ref_annotations)]  # type: ignore
                if self._incr_ref(ref, annotated):
                    return
        return super().annotated(tp, annotations)

    def any(self):
        pass

    def collection(self, cls: Type[Collection], value_type: AnyType):
        self.visit(value_type)

    def enum(self, cls: Type[Enum]):
        pass

    def literal(self, values: Sequence[Any]):
        pass

    def mapping(self, cls: Type[Mapping], key_type: AnyType, value_type: AnyType):
        self.visit(key_type)
        self.visit(value_type)

    def object(self, tp: AnyType, fields: Sequence[ObjectField]):
        for field in fields:
            self.visit_with_conv(field.type, self._field_conversion(field))

    def primitive(self, cls: Type):
        pass

    def tuple(self, types: Sequence[AnyType]):
        for cls in types:
            self.visit(cls)

    def _visited_union(self, results: Sequence):
        pass

    def visit_conversion(
        self,
        tp: AnyType,
        conversion: Optional[Any],
        dynamic: bool,
        next_conversion: Optional[AnyConversion] = None,
    ):
        ref_types = []
        if not dynamic:
            for ref_tp in self.resolve_conversion(tp):
                ref_types.append(ref_tp)
                if self._incr_ref(get_type_name(ref_tp).json_schema, ref_tp):
                    return
        if not is_hashable(tp):
            return super().visit_conversion(tp, conversion, dynamic, next_conversion)
        # 2 because the first type encountered of the recursive cycle can have no ref
        # (see test_recursive_by_conversion_schema)
        if self._rec_guard[(tp, self._conversion)] > 2:
            raise TypeError(f"Recursive type {tp} need a ref")
        self._rec_guard[(tp, self._conversion)] += 1
        try:
            super().visit_conversion(tp, conversion, dynamic, next_conversion)
        except Unsupported:
            for ref_tp in ref_types:
                self.refs.pop(get_type_name(ref_tp).json_schema, ...)  # type: ignore
        finally:
            self._rec_guard[(tp, self._conversion)] -= 1


class DeserializationRefsExtractor(
    RefsExtractor, DeserializationVisitor, DeserializationObjectVisitor
):
    pass


class SerializationRefsExtractor(
    RefsExtractor, SerializationVisitor, SerializationObjectVisitor
):
    pass



from collections import defaultdict
from dataclasses import dataclass
from enum import Enum
from functools import lru_cache
from inspect import Parameter, signature
from typing import (
    Any,
    Awaitable,
    Callable,
    Collection,
    Dict,
    Iterator,
    Mapping,
    MutableMapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    overload,
)

import graphql

from apischema import UndefinedType
from apischema.aliases import Aliaser
from apischema.cache import CacheAwareDict, cache
from apischema.conversions import Conversion
from apischema.conversions.conversions import AnyConversion, DefaultConversion
from apischema.deserialization import deserialization_method
from apischema.methods import method_registerer
from apischema.objects import ObjectField
from apischema.ordering import Ordering
from apischema.schemas import Schema
from apischema.serialization import (
    PassThroughOptions,
    SerializationMethod,
    SerializationMethodVisitor,
)
from apischema.serialization.serialized_methods import (
    ErrorHandler,
    SerializedMethod,
    _get_methods,
    serialized as register_serialized,
)
from apischema.types import AnyType, NoneType, Undefined
from apischema.typing import is_type
from apischema.utils import (
    awaitable_origin,
    deprecate_kwargs,
    empty_dict,
    get_args2,
    get_origin_or_type2,
    identity,
    is_async,
    is_union_of,
    keep_annotations,
)
from apischema.validation.errors import ValidationError


class PartialSerializationMethodVisitor(SerializationMethodVisitor):
    use_cache = False

    @property
    def _factory(self) -> Callable[[type], SerializationMethod]:
        return lambda _: identity

    def enum(self, cls: Type[Enum]) -> SerializationMethod:
        return identity

    def object(self, tp: AnyType, fields: Sequence[ObjectField]) -> SerializationMethod:
        return identity

    def visit(self, tp: AnyType) -> SerializationMethod:
        if tp is UndefinedType:
            return lambda obj: None
        return super().visit(tp)


@cache
def partial_serialization_method_factory(
    aliaser: Aliaser,
    conversion: Optional[AnyConversion],
    default_conversion: DefaultConversion,
) -> Callable[[AnyType], SerializationMethod]:
    @lru_cache()
    def factory(tp: AnyType) -> SerializationMethod:
        return PartialSerializationMethodVisitor(
            additional_properties=False,
            aliaser=aliaser,
            check_type=False,
            default_conversion=default_conversion,
            exclude_defaults=False,
            exclude_none=False,
            exclude_unset=False,
            fall_back_on_any=False,
            pass_through_options=PassThroughOptions(),
        ).visit_with_conv(tp, conversion)

    return factory


def unwrap_awaitable(tp: AnyType) -> AnyType:
    if get_origin_or_type2(tp) == awaitable_origin:
        return keep_annotations(get_args2(tp)[0] if get_args2(tp) else Any, tp)
    else:
        return tp


@dataclass(frozen=True)
class Resolver(SerializedMethod):
    parameters: Sequence[Parameter]
    parameters_metadata: Mapping[str, Mapping]

    def error_type(self) -> AnyType:
        return unwrap_awaitable(super().error_type())

    def return_type(self, return_type: AnyType) -> AnyType:
        return super().return_type(unwrap_awaitable(return_type))


_resolvers: MutableMapping[Type, Dict[str, Resolver]] = CacheAwareDict(
    defaultdict(dict)
)


def get_resolvers(tp: AnyType) -> Collection[Tuple[Resolver, Mapping[str, AnyType]]]:
    return _get_methods(tp, _resolvers)


def none_error_handler(
    __error: Exception, __obj: Any, __info: graphql.GraphQLResolveInfo, **kwargs
) -> None:
    return None


def resolver_parameters(
    resolver: Callable, *, check_first: bool
) -> Iterator[Parameter]:
    first = True
    for param in signature(resolver).parameters.values():
        if param.kind is Parameter.POSITIONAL_ONLY:
            raise TypeError("Resolver can not have positional only parameters")
        if param.kind in {Parameter.POSITIONAL_OR_KEYWORD, Parameter.KEYWORD_ONLY}:
            if param.annotation is Parameter.empty and (check_first or not first):
                raise TypeError("Resolver parameters must be typed")
            yield param
        first = False


MethodOrProp = TypeVar("MethodOrProp", Callable, property)


@overload
def resolver(__method_or_property: MethodOrProp) -> MethodOrProp:
    ...


@overload
def resolver(
    alias: str = None,
    *,
    conversion: AnyConversion = None,
    error_handler: ErrorHandler = Undefined,
    order: Optional[Ordering] = None,
    schema: Schema = None,
    parameters_metadata: Mapping[str, Mapping] = None,
    serialized: bool = False,
    owner: Type = None,
) -> Callable[[MethodOrProp], MethodOrProp]:
    ...


@deprecate_kwargs({"conversions": "conversion"})
def resolver(
    __arg=None,
    *,
    alias: str = None,
    conversion: AnyConversion = None,
    error_handler: ErrorHandler = Undefined,
    order: Optional[Ordering] = None,
    schema: Schema = None,
    parameters_metadata: Mapping[str, Mapping] = None,
    serialized: bool = False,
    owner: Type = None,
):
    def register(func: Callable, owner: Type, alias2: str):
        alias2 = alias or alias2
        _, *parameters = resolver_parameters(func, check_first=owner is None)
        error_handler2 = error_handler
        if error_handler2 is None:
            error_handler2 = none_error_handler
        elif error_handler2 is Undefined:
            error_handler2 = None
        resolver = Resolver(
            func,
            alias2,
            conversion,
            error_handler2,
            order,
            schema,
            parameters,
            parameters_metadata or {},
        )
        _resolvers[owner][alias2] = resolver
        if serialized:
            if is_async(func):
                raise TypeError("Async resolver cannot be used as a serialized method")
            try:
                register_serialized(
                    alias=alias2,
                    conversion=conversion,
                    schema=schema,
                    error_handler=error_handler,
                    owner=owner,
                )(func)
            except Exception:
                raise TypeError("Resolver cannot be used as a serialized method")

    if isinstance(__arg, str):
        alias = __arg
        __arg = None
    return method_registerer(__arg, owner, register)


T = TypeVar("T")
U = TypeVar("U")


def as_async(func: Callable[[T], U]) -> Callable[[Awaitable[T]], Awaitable[U]]:
    async def wrapper(arg: Awaitable[T]) -> U:
        return func(await arg)

    return wrapper


def resolver_resolve(
    resolver: Resolver,
    types: Mapping[str, AnyType],
    aliaser: Aliaser,
    default_deserialization: DefaultConversion,
    default_serialization: DefaultConversion,
    serialized: bool = True,
) -> Callable:
    # graphql deserialization will give Enum objects instead of strings
    def handle_enum(tp: AnyType) -> Optional[AnyConversion]:
        if is_type(tp) and issubclass(tp, Enum):
            return Conversion(identity, source=Any, target=tp)
        return default_deserialization(tp)

    parameters, info_parameter = [], None
    for param in resolver.parameters:
        param_type = types[param.name]
        if is_union_of(param_type, graphql.GraphQLResolveInfo):
            info_parameter = param.name
        else:
            param_field = ObjectField(
                param.name,
                param_type,
                param.default is Parameter.empty,
                resolver.parameters_metadata.get(param.name, empty_dict),
                param.default,
            )
            deserializer = deserialization_method(
                param_type,
                additional_properties=False,
                aliaser=aliaser,
                coerce=False,
                conversion=param_field.deserialization,
                default_conversion=handle_enum,
                fall_back_on_default=False,
                schema=param_field.schema,
            )
            opt_param = is_union_of(param_type, NoneType) or param.default is None
            parameters.append(
                (
                    aliaser(param_field.alias),
                    param.name,
                    deserializer,
                    opt_param,
                    param_field.required,
                )
            )
    func, error_handler = resolver.func, resolver.error_handler
    method_factory = partial_serialization_method_factory(
        aliaser, resolver.conversion, default_serialization
    )

    serialize_result: Callable[[Any], Any]
    if not serialized:
        serialize_result = identity
    elif is_async(resolver.func):
        serialize_result = as_async(method_factory(types["return"]))
    else:
        serialize_result = method_factory(types["return"])
    serialize_error: Optional[Callable[[Any], Any]]
    if error_handler is None:
        serialize_error = None
    elif is_async(error_handler):
        serialize_error = as_async(method_factory(resolver.error_type()))
    else:
        serialize_error = method_factory(resolver.error_type())

    def resolve(__self, __info, **kwargs):
        values = {}
        errors: Dict[str, ValidationError] = {}
        for alias, param_name, deserializer, opt_param, required in parameters:
            if alias in kwargs:
                # It is possible for the parameter to be non-optional in Python
                # type hints but optional in the generated schema. In this case
                # we should ignore it.
                # See: https://github.com/wyfo/apischema/pull/130#issuecomment-845497392
                if not opt_param and kwargs[alias] is None:
                    assert not required
                    continue
                try:
                    values[param_name] = deserializer(kwargs[alias])
                except ValidationError as err:
                    errors[aliaser(param_name)] = err
            elif opt_param and required:
                values[param_name] = None

        if errors:
            raise ValueError(ValidationError(children=errors).errors)
        if info_parameter:
            values[info_parameter] = __info
        try:
            return serialize_result(func(__self, **values))
        except Exception as error:
            if error_handler is None:
                raise
            assert serialize_error is not None
            return serialize_error(error_handler(error, __self, __info, **kwargs))

    return resolve



from collections import defaultdict
from dataclasses import dataclass, replace
from enum import Enum
from functools import lru_cache
from typing import (
    AbstractSet,
    Any,
    Callable,
    Collection,
    Dict,
    List,
    Mapping,
    Optional,
    Pattern,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    overload,
)

from apischema.aliases import Aliaser
from apischema.cache import cache
from apischema.conversions.conversions import AnyConversion, DefaultConversion
from apischema.conversions.visitor import (
    Deserialization,
    DeserializationVisitor,
    sub_conversion,
)
from apischema.dependencies import get_dependent_required
from apischema.deserialization.coercion import Coerce, Coercer
from apischema.deserialization.flattened import get_deserialization_flattened_aliases
from apischema.json_schema.patterns import infer_pattern
from apischema.json_schema.types import bad_type
from apischema.metadata.implem import ValidatorsMetadata
from apischema.metadata.keys import SCHEMA_METADATA, VALIDATORS_METADATA
from apischema.objects import ObjectField
from apischema.objects.fields import FieldKind
from apischema.objects.visitor import DeserializationObjectVisitor
from apischema.recursion import RecursiveConversionsVisitor
from apischema.schemas import Schema, get_schema
from apischema.schemas.constraints import Check, Constraints, merge_constraints
from apischema.types import AnyType, NoneType
from apischema.typing import get_args, get_origin
from apischema.utils import (
    Lazy,
    PREFIX,
    deprecate_kwargs,
    get_origin_or_type,
    literal_values,
    opt_or,
)
from apischema.validation import get_validators
from apischema.validation.errors import ErrorKey, ValidationError, merge_errors
from apischema.validation.mock import ValidatorMock
from apischema.validation.validators import Validator, validate
from apischema.visitor import Unsupported

MISSING_PROPERTY = "missing property"
UNEXPECTED_PROPERTY = "unexpected property"

NOT_NONE = object()

INIT_VARS_ATTR = f"{PREFIX}_init_vars"

T = TypeVar("T")


DeserializationMethod = Callable[[Any], T]
Factory = Callable[[Optional[Constraints], Sequence[Validator]], DeserializationMethod]


@dataclass(frozen=True)
class DeserializationMethodFactory:
    factory: Factory
    cls: Optional[type] = None
    constraints: Optional[Constraints] = None
    validators: Tuple[Validator, ...] = ()

    def merge(
        self, constraints: Optional[Constraints], validators: Sequence[Validator] = ()
    ) -> "DeserializationMethodFactory":
        if constraints is None and not validators:
            return self
        return replace(
            self,
            constraints=merge_constraints(self.constraints, constraints),
            validators=(*validators, *self.validators),
        )

    @property  # type: ignore
    @lru_cache()
    def method(self) -> DeserializationMethod:
        return self.factory(self.constraints, self.validators)  # type: ignore


def get_constraints(schema: Optional[Schema]) -> Optional[Constraints]:
    return schema.constraints if schema is not None else None


def get_constraint_checks(
    constraints: Optional[Constraints], cls: type
) -> Collection[Tuple[Check, Any, str]]:
    return () if constraints is None else constraints.checks_by_type[cls]


class DeserializationMethodVisitor(
    RecursiveConversionsVisitor[Deserialization, DeserializationMethodFactory],
    DeserializationVisitor[DeserializationMethodFactory],
    DeserializationObjectVisitor[DeserializationMethodFactory],
):
    def __init__(
        self,
        additional_properties: bool,
        aliaser: Aliaser,
        coercer: Optional[Coercer],
        default_conversion: DefaultConversion,
        fall_back_on_default: bool,
    ):
        super().__init__(default_conversion)
        self.additional_properties = additional_properties
        self.aliaser = aliaser
        self.coercer = coercer
        self.fall_back_on_default = fall_back_on_default

    def _recursive_result(
        self, lazy: Lazy[DeserializationMethodFactory]
    ) -> DeserializationMethodFactory:
        def factory(
            constraints: Optional[Constraints], validators: Sequence[Validator]
        ) -> DeserializationMethod:
            rec_method = None

            def method(data: Any) -> Any:
                nonlocal rec_method
                if rec_method is None:
                    rec_method = lazy().merge(constraints, validators).method
                return rec_method(data)

            return method

        return DeserializationMethodFactory(factory)

    def visit_not_recursive(self, tp: AnyType) -> DeserializationMethodFactory:
        return deserialization_method_factory(
            tp,
            self.additional_properties,
            self.aliaser,
            self.coercer,
            self._conversion,
            self.default_conversion,
            self.fall_back_on_default,
        )

    def annotated(
        self, tp: AnyType, annotations: Sequence[Any]
    ) -> DeserializationMethodFactory:
        factory = super().annotated(tp, annotations)
        for annotation in reversed(annotations):
            if isinstance(annotation, Mapping):
                factory = factory.merge(
                    get_constraints(annotation.get(SCHEMA_METADATA)),
                    annotation.get(
                        VALIDATORS_METADATA, ValidatorsMetadata(())
                    ).validators,
                )
        return factory

    def _factory(
        self, factory: Factory, cls: Optional[type] = None, validation: bool = True
    ) -> DeserializationMethodFactory:
        def wrapper(
            constraints: Optional[Constraints], validators: Sequence[Validator]
        ) -> DeserializationMethod:
            method: DeserializationMethod
            if validation and validators:
                wrapped, aliaser = factory(constraints, ()), self.aliaser

                def method(data: Any) -> Any:
                    result = wrapped(data)
                    validate(result, validators, aliaser=aliaser)
                    return result

            else:
                method = factory(constraints, validators)
            if self.coercer is not None and cls is not None:
                coercer = self.coercer

                def wrapper(data: Any) -> Any:
                    assert cls is not None
                    return method(coercer(cls, data))

                return wrapper

            else:
                return method

        return DeserializationMethodFactory(wrapper, cls)

    def any(self) -> DeserializationMethodFactory:
        def factory(constraints: Optional[Constraints], _) -> DeserializationMethod:
            checks = None if constraints is None else constraints.checks_by_type

            def method(data: Any) -> Any:
                if checks is not None:
                    if data.__class__ in checks:
                        errors = [
                            err
                            for check, attr, err in checks[data.__class__]
                            if check(data, attr)
                        ]
                        if errors:
                            raise ValidationError(errors)
                return data

            return method

        return self._factory(factory)

    def collection(
        self, cls: Type[Collection], value_type: AnyType
    ) -> DeserializationMethodFactory:
        value_factory = self.visit(value_type)

        def factory(constraints: Optional[Constraints], _) -> DeserializationMethod:
            deserialize_value = value_factory.method
            checks = get_constraint_checks(constraints, list)
            constructor: Optional[Callable[[list], Collection]] = None
            if issubclass(cls, AbstractSet):
                constructor = set
            elif issubclass(cls, tuple):
                constructor = tuple

            def method(data: Any) -> Any:
                if not isinstance(data, list):
                    raise bad_type(data, list)
                elt_errors: Dict[ErrorKey, ValidationError] = {}
                values: list = [None] * len(data)
                index = 0  # don't use `enumerate` for performance
                for elt in data:
                    try:
                        values[index] = deserialize_value(elt)
                    except ValidationError as err:
                        elt_errors[index] = err
                    index += 1
                if checks:
                    errors = [err for check, attr, err in checks if check(data, attr)]
                    if errors or elt_errors:
                        raise ValidationError(errors, elt_errors)
                elif elt_errors:
                    raise ValidationError([], elt_errors)
                return constructor(values) if constructor else values

            return method

        return self._factory(factory, list)

    def enum(self, cls: Type[Enum]) -> DeserializationMethodFactory:
        return self.literal(list(cls))

    def literal(self, values: Sequence[Any]) -> DeserializationMethodFactory:
        def factory(constraints: Optional[Constraints], _) -> DeserializationMethod:
            value_map = dict(zip(literal_values(values), values))
            types = list(set(map(type, value_map))) if self.coercer else []
            error = f"not one of {list(value_map)}"
            coercer = self.coercer

            def method(data: Any) -> Any:
                try:
                    return value_map[data]
                except KeyError:
                    if coercer:
                        for cls in types:
                            try:
                                return value_map[coercer(cls, data)]
                            except IndexError:
                                pass
                    raise ValidationError([error])

            return method

        return self._factory(factory)

    def mapping(
        self, cls: Type[Mapping], key_type: AnyType, value_type: AnyType
    ) -> DeserializationMethodFactory:
        key_factory, value_factory = self.visit(key_type), self.visit(value_type)

        def factory(constraints: Optional[Constraints], _) -> DeserializationMethod:
            deserialize_key = key_factory.method
            deserialize_value = value_factory.method
            checks = get_constraint_checks(constraints, dict)

            def method(data: Any) -> Any:
                if not isinstance(data, dict):
                    raise bad_type(data, dict)
                item_errors: Dict[ErrorKey, ValidationError] = {}
                items = {}
                for key, value in data.items():
                    assert isinstance(key, str)
                    try:
                        items[deserialize_key(key)] = deserialize_value(value)
                    except ValidationError as err:
                        item_errors[key] = err
                if checks:
                    errors = [err for check, attr, err in checks if check(data, attr)]
                    if errors or item_errors:
                        raise ValidationError(errors, item_errors)
                elif item_errors:
                    raise ValidationError([], item_errors)
                return items

            return method

        return self._factory(factory, dict)

    def object(
        self, tp: Type, fields: Sequence[ObjectField]
    ) -> DeserializationMethodFactory:
        field_factories = [
            self.visit_with_conv(f.type, f.deserialization).merge(
                get_constraints(f.schema), f.validators
            )
            for f in fields
        ]

        def factory(
            constraints: Optional[Constraints], validators: Sequence[Validator]
        ) -> DeserializationMethod:
            cls = get_origin_or_type(tp)
            alias_by_name = {field.name: self.aliaser(field.alias) for field in fields}
            requiring: Dict[str, Set[str]] = defaultdict(set)
            for f, reqs in get_dependent_required(cls).items():
                for req in reqs:
                    requiring[req].add(alias_by_name[f])
            normal_fields, flattened_fields, pattern_fields = [], [], []
            additional_field = None
            for field, field_factory in zip(fields, field_factories):
                deserialize_field: DeserializationMethod = field_factory.method
                fall_back_on_default = (
                    field.fall_back_on_default or self.fall_back_on_default
                )
                if field.flattened:
                    flattened_aliases = get_deserialization_flattened_aliases(
                        cls, field, self.default_conversion
                    )
                    flattened_fields.append(
                        (
                            field.name,
                            set(map(self.aliaser, flattened_aliases)),
                            deserialize_field,
                            fall_back_on_default,
                        )
                    )
                elif field.pattern_properties is not None:
                    field_pattern = field.pattern_properties
                    if field_pattern is ...:
                        field_pattern = infer_pattern(
                            field.type, self.default_conversion
                        )
                    assert isinstance(field_pattern, Pattern)
                    pattern_fields.append(
                        (
                            field.name,
                            field_pattern,
                            deserialize_field,
                            fall_back_on_default,
                        )
                    )
                elif field.additional_properties:
                    additional_field = (
                        field.name,
                        deserialize_field,
                        fall_back_on_default,
                    )
                else:
                    normal_fields.append(
                        (
                            field.name,
                            self.aliaser(field.alias),
                            deserialize_field,
                            field.required,
                            requiring[field.name],
                            fall_back_on_default,
                        )
                    )
            has_aggregate_field = (
                flattened_fields or pattern_fields or (additional_field is not None)
            )
            post_init_modified = {field.name for field in fields if field.post_init}
            checks = get_constraint_checks(constraints, dict)
            aliaser = self.aliaser
            additional_properties = self.additional_properties
            all_aliases = set(alias_by_name.values())
            init_defaults = [
                (f.name, f.default_factory)
                for f in fields
                if f.kind == FieldKind.WRITE_ONLY
            ]

            def method(data: Any) -> Any:
                if not isinstance(data, dict):
                    raise bad_type(data, dict)
                values: Dict[str, Any] = {}
                fields_count = 0
                errors = (
                    [err for check, attr, err in checks if check(data, attr)]
                    if checks
                    else []
                )
                field_errors: Dict[ErrorKey, ValidationError] = {}
                for (
                    name,
                    alias,
                    deserialize_field,
                    required,
                    required_by,
                    fall_back_on_default,
                ) in normal_fields:
                    if required:
                        try:
                            value = data[alias]
                        except KeyError:
                            field_errors[alias] = ValidationError([MISSING_PROPERTY])
                        else:
                            fields_count += 1
                            try:
                                values[name] = deserialize_field(value)
                            except ValidationError as err:
                                field_errors[alias] = err
                    elif alias in data:
                        fields_count += 1
                        try:
                            values[name] = deserialize_field(data[alias])
                        except ValidationError as err:
                            if not fall_back_on_default:
                                field_errors[alias] = err
                    elif required_by and not required_by.isdisjoint(data):
                        requiring = sorted(required_by & data.keys())
                        msg = f"missing property (required by {requiring})"
                        field_errors[alias] = ValidationError([msg])
                if has_aggregate_field:
                    remain = data.keys() - all_aliases
                    for (
                        name,
                        flattened_alias,
                        deserialize_field,
                        fall_back_on_default,
                    ) in flattened_fields:
                        flattened = {
                            alias: data[alias]
                            for alias in flattened_alias
                            if alias in data
                        }
                        remain.difference_update(flattened)
                        try:
                            values[name] = deserialize_field(flattened)
                        except ValidationError as err:
                            if not fall_back_on_default:
                                errors.extend(err.messages)
                                field_errors.update(err.children)
                    for (
                        name,
                        pattern,
                        deserialize_field,
                        fall_back_on_default,
                    ) in pattern_fields:
                        matched = {
                            key: data[key] for key in remain if pattern.match(key)
                        }
                        remain.difference_update(matched)
                        try:
                            values[name] = deserialize_field(matched)
                        except ValidationError as err:
                            if not fall_back_on_default:
                                errors.extend(err.messages)
                                field_errors.update(err.children)
                    if additional_field:
                        name, deserialize_field, fall_back_on_default = additional_field
                        additional = {key: data[key] for key in remain}
                        try:
                            values[name] = deserialize_field(additional)
                        except ValidationError as err:
                            if not fall_back_on_default:
                                errors.extend(err.messages)
                                field_errors.update(err.children)
                    elif remain and not additional_properties:
                        for key in remain:
                            field_errors[key] = ValidationError([UNEXPECTED_PROPERTY])
                elif not additional_properties and len(data) != fields_count:
                    for key in data.keys() - all_aliases:
                        field_errors[key] = ValidationError([UNEXPECTED_PROPERTY])
                validators2: Sequence[Validator]
                if validators:
                    init: Dict[str, Any] = {}
                    for name, default_factory in init_defaults:
                        if name in values:
                            init[name] = values[name]
                        elif name not in field_errors:
                            assert default_factory is not None
                            init[name] = default_factory()
                    # Don't keep validators when all dependencies are default
                    validators2 = [
                        v
                        for v in validators
                        if not v.dependencies.isdisjoint(values.keys())
                    ]
                    if field_errors or errors:
                        error = ValidationError(errors, field_errors)
                        invalid_fields = field_errors.keys() | post_init_modified
                        try:
                            validate(
                                ValidatorMock(cls, values),
                                [
                                    v
                                    for v in validators2
                                    if v.dependencies.isdisjoint(invalid_fields)
                                ],
                                init,
                                aliaser=aliaser,
                            )
                        except ValidationError as err:
                            error = merge_errors(error, err)
                        raise error
                elif field_errors or errors:
                    raise ValidationError(errors, field_errors)
                else:
                    validators2, init = (), ...  # type: ignore # only for linter
                try:
                    res = cls(**values)
                except (AssertionError, ValidationError):
                    raise
                except TypeError as err:
                    if str(err).startswith("__init__() got"):
                        raise Unsupported(cls)
                    else:
                        raise ValidationError([str(err)])
                except Exception as err:
                    raise ValidationError([str(err)])
                if validators:
                    validate(res, validators2, init, aliaser=aliaser)
                return res

            return method

        return self._factory(factory, dict, validation=False)

    def primitive(self, cls: Type) -> DeserializationMethodFactory:
        def factory(constraints: Optional[Constraints], _) -> DeserializationMethod:
            checks = get_constraint_checks(constraints, cls)
            if cls is NoneType:

                def method(data: Any) -> Any:
                    if data is not None:
                        raise bad_type(data, cls)
                    return data

            elif cls is not float and not checks:

                def method(data: Any) -> Any:
                    if not isinstance(data, cls):
                        raise bad_type(data, cls)
                    return data

            elif cls is not float and len(checks) == 1:
                ((check, attr, err),) = checks

                def method(data: Any) -> Any:
                    if not isinstance(data, cls):
                        raise bad_type(data, cls)
                    elif check(data, attr):
                        raise ValidationError([err])
                    return data

            else:
                is_float = cls is float

                def method(data: Any) -> Any:
                    if not isinstance(data, cls):
                        if is_float and isinstance(data, int):
                            data = float(data)
                        else:
                            raise bad_type(data, cls)
                    if checks:
                        errors = [
                            err for check, attr, err in checks if check(data, attr)
                        ]
                        if errors:
                            raise ValidationError(errors)
                    return data

            return method

        return self._factory(factory, cls)

    def subprimitive(self, cls: Type, superclass: Type) -> DeserializationMethodFactory:
        primitive_factory = self.primitive(superclass)

        def factory(
            constraints: Optional[Constraints], validators: Sequence[Validator]
        ) -> DeserializationMethod:
            deserialize_primitive = primitive_factory.merge(
                constraints, validators
            ).method

            def method(data: Any) -> Any:
                return superclass(deserialize_primitive(data))

            return method

        return replace(primitive_factory, factory=factory)

    def tuple(self, types: Sequence[AnyType]) -> DeserializationMethodFactory:
        elt_factories = [self.visit(tp) for tp in types]

        def factory(constraints: Optional[Constraints], _) -> DeserializationMethod:
            expected_len = len(types)
            (_, _, min_err), (_, _, max_err) = Constraints(
                min_items=len(types), max_items=len(types)
            ).checks_by_type[list]
            elt_methods = list(enumerate(fact.method for fact in elt_factories))
            checks = get_constraint_checks(constraints, list)

            def method(data: Any) -> Any:
                if not isinstance(data, list):
                    raise bad_type(data, list)
                if len(data) != expected_len:
                    raise ValidationError([min_err, max_err])
                elt_errors: Dict[ErrorKey, ValidationError] = {}
                elts: List[Any] = [None] * expected_len
                for i, deserialize_elt in elt_methods:
                    try:
                        elts[i] = deserialize_elt(data[i])
                    except ValidationError as err:
                        elt_errors[i] = err
                if checks:
                    errors = [err for check, attr, err in checks if check(data, attr)]
                    if errors or elt_errors:
                        raise ValidationError(errors, elt_errors)
                elif elt_errors:
                    raise ValidationError([], elt_errors)
                return tuple(elts)

            return method

        return self._factory(factory, list)

    def union(self, alternatives: Sequence[AnyType]) -> DeserializationMethodFactory:
        alt_factories = self._union_results(alternatives)
        if len(alt_factories) == 1:
            return alt_factories[0]

        def factory(constraints: Optional[Constraints], _) -> DeserializationMethod:
            alt_methods = [fact.merge(constraints).method for fact in alt_factories]
            # method_by_cls cannot replace alt_methods, because there could be several
            # methods for one class
            method_by_cls = dict(zip((f.cls for f in alt_factories), alt_methods))
            if NoneType in alternatives and len(alt_methods) == 2:
                deserialize_alt = next(
                    meth
                    for fact, meth in zip(alt_factories, alt_methods)
                    if fact.cls is not NoneType
                )
                coercer = self.coercer

                def method(data: Any) -> Any:
                    if data is None:
                        return None
                    try:
                        return deserialize_alt(data)
                    except ValidationError as err:
                        if coercer and coercer(NoneType, data) is None:
                            return None
                        else:
                            raise merge_errors(err, bad_type(data, NoneType))

            elif None not in method_by_cls and len(method_by_cls) == len(alt_factories):
                classes = tuple(cls for cls in method_by_cls if cls is not None)

                def method(data: Any) -> Any:
                    try:
                        return method_by_cls[data.__class__](data)
                    except KeyError:
                        raise bad_type(data, *classes) from None
                    except ValidationError as err:
                        other_classes = (
                            cls for cls in classes if cls is not data.__class__
                        )
                        raise merge_errors(err, bad_type(data, *other_classes))

            else:

                def method(data: Any) -> Any:
                    error = None
                    for deserialize_alt in alt_methods:
                        try:
                            return deserialize_alt(data)
                        except ValidationError as err:
                            error = merge_errors(error, err)
                    assert error is not None
                    raise error

            return method

        return self._factory(factory)

    def _visit_conversion(
        self,
        tp: AnyType,
        conversion: Deserialization,
        dynamic: bool,
        next_conversion: Optional[AnyConversion],
    ) -> DeserializationMethodFactory:
        assert conversion
        conv_factories = [
            self.visit_with_conv(conv.source, sub_conversion(conv, next_conversion))
            for conv in conversion
        ]

        def factory(constraints: Optional[Constraints], _) -> DeserializationMethod:
            conv_methods = [
                ((fact if dynamic else fact.merge(constraints)).method, conv.converter)
                for conv, fact in zip(conversion, conv_factories)
            ]
            method: DeserializationMethod
            if len(conv_methods) == 1:
                deserialize_alt, converter = conv_methods[0]

                def method(data: Any) -> Any:
                    try:
                        return converter(deserialize_alt(data))
                    except (ValidationError, AssertionError):
                        raise
                    except Exception as err:
                        raise ValidationError([str(err)])

            else:

                def method(data: Any) -> Any:
                    error: Optional[ValidationError] = None
                    for deserialize_alt, converter in conv_methods:
                        try:
                            value = deserialize_alt(data)
                        except ValidationError as err:
                            error = merge_errors(error, err)
                        else:
                            try:
                                return converter(value)
                            except (ValidationError, AssertionError):
                                raise
                            except Exception as err:
                                raise ValidationError([str(err)])
                    assert error is not None
                    raise error

            return method

        return self._factory(factory, validation=not dynamic)

    def visit_conversion(
        self,
        tp: AnyType,
        conversion: Optional[Deserialization],
        dynamic: bool,
        next_conversion: Optional[AnyConversion] = None,
    ) -> DeserializationMethodFactory:
        factory = super().visit_conversion(tp, conversion, dynamic, next_conversion)
        if not dynamic:
            factory = factory.merge(get_constraints(get_schema(tp)), get_validators(tp))
            if get_args(tp):
                factory = factory.merge(
                    get_constraints(get_schema(get_origin(tp))),
                    get_validators(get_origin(tp)),
                )
        return factory


@cache
def deserialization_method_factory(
    tp: AnyType,
    additional_properties: bool,
    aliaser: Aliaser,
    coercer: Optional[Coercer],
    conversion: Optional[AnyConversion],
    default_conversion: DefaultConversion,
    fall_back_on_default: bool,
) -> DeserializationMethodFactory:
    return DeserializationMethodVisitor(
        additional_properties,
        aliaser,
        coercer,
        default_conversion,
        fall_back_on_default,
    ).visit_with_conv(tp, conversion)


@overload
def deserialization_method(
    type: Type[T],
    *,
    additional_properties: bool = None,
    aliaser: Aliaser = None,
    coerce: Coerce = None,
    conversion: AnyConversion = None,
    default_conversion: DefaultConversion = None,
    fall_back_on_default: bool = None,
    schema: Schema = None,
) -> DeserializationMethod[T]:
    ...


@overload
def deserialization_method(
    type: AnyType,
    *,
    additional_properties: bool = None,
    aliaser: Aliaser = None,
    coerce: Coerce = None,
    conversion: AnyConversion = None,
    default_conversion: DefaultConversion = None,
    fall_back_on_default: bool = None,
    schema: Schema = None,
) -> DeserializationMethod:
    ...


def deserialization_method(
    type: AnyType,
    *,
    additional_properties: bool = None,
    aliaser: Aliaser = None,
    coerce: Coerce = None,
    conversion: AnyConversion = None,
    default_conversion: DefaultConversion = None,
    fall_back_on_default: bool = None,
    schema: Schema = None,
) -> DeserializationMethod:
    from apischema import settings

    coercer: Optional[Coercer] = None
    if callable(coerce):
        coercer = coerce
    elif opt_or(coerce, settings.deserialization.coerce):
        coercer = settings.deserialization.coercer
    return (
        deserialization_method_factory(
            type,
            opt_or(additional_properties, settings.additional_properties),
            opt_or(aliaser, settings.aliaser),
            coercer,
            conversion,
            opt_or(default_conversion, settings.deserialization.default_conversion),
            opt_or(fall_back_on_default, settings.deserialization.fall_back_on_default),
        )
        .merge(get_constraints(schema), ())
        .method
    )


@overload
def deserialize(
    type: Type[T],
    data: Any,
    *,
    additional_properties: bool = None,
    aliaser: Aliaser = None,
    coerce: Coerce = None,
    conversion: AnyConversion = None,
    default_conversion: DefaultConversion = None,
    fall_back_on_default: bool = None,
    schema: Schema = None,
) -> T:
    ...


@overload
def deserialize(
    type: AnyType,
    data: Any,
    *,
    additional_properties: bool = None,
    aliaser: Aliaser = None,
    coerce: Coerce = None,
    conversion: AnyConversion = None,
    default_conversion: DefaultConversion = None,
    fall_back_on_default: bool = None,
    schema: Schema = None,
) -> Any:
    ...


@deprecate_kwargs(
    {
        "coercion": "coerce",
        "conversions": "conversion",
        "default_fallback": "fall_back_on_default",
    }
)
def deserialize(
    type: AnyType,
    data: Any,
    *,
    additional_properties: bool = None,
    aliaser: Aliaser = None,
    coerce: Coerce = None,
    conversion: AnyConversion = None,
    default_conversion: DefaultConversion = None,
    fall_back_on_default: bool = None,
    schema: Schema = None,
) -> Any:
    return deserialization_method(
        type,
        additional_properties=additional_properties,
        aliaser=aliaser,
        coerce=coerce,
        conversion=conversion,
        default_conversion=default_conversion,
        fall_back_on_default=fall_back_on_default,
        schema=schema,
    )(data)



import ast
import inspect
import textwrap
from typing import AbstractSet, Callable, Collection, Dict, Set


Dependencies = AbstractSet[str]


class DependencyFinder(ast.NodeVisitor):
    def __init__(self, param: str):
        self.param = param
        self.dependencies: Set[str] = set()

    def visit_Attribute(self, node):
        self.generic_visit(node)
        if isinstance(node.value, ast.Name) and node.value.id == self.param:
            self.dependencies.add(node.attr)

    # TODO Add warning in case of function call with self in parameter
    # or better, follow the call, but it would be too hard (local import, etc.)


def first_parameter(func: Callable) -> str:
    try:
        return next(iter(inspect.signature(func).parameters))
    except StopIteration:
        raise TypeError("Cannot compute dependencies if no parameter")


def find_dependencies(func: Callable) -> Dependencies:
    try:
        finder = DependencyFinder(first_parameter(func))
        finder.visit(ast.parse(textwrap.dedent(inspect.getsource(func))))
    except ValueError:
        return set()
    return finder.dependencies


cache: Dict[Callable, Dependencies] = {}


def find_all_dependencies(
    cls: type, func: Callable, rec_guard: Collection[str] = ()
) -> Dependencies:
    """Dependencies contains class variables (because they can be "fake" ones as in
    dataclasses)"""
    if func not in cache:
        dependencies = set(find_dependencies(func))
        for attr in list(dependencies):
            if not hasattr(cls, attr):
                continue
            member = getattr(cls, attr)
            if isinstance(member, property):
                member = member.fget
            if callable(member):
                dependencies.remove(attr)
                if member in rec_guard:
                    continue
                rec_deps = find_all_dependencies(cls, member, {*rec_guard, member})
                dependencies.update(rec_deps)
        cache[func] = dependencies
    return cache[func]



from dataclasses import dataclass, field as field_, replace
from enum import Enum
from functools import wraps
from inspect import Parameter, iscoroutinefunction
from itertools import chain
from typing import (
    Any,
    AsyncIterable,
    AsyncIterator,
    Callable,
    Collection,
    Dict,
    Generic,
    Iterable,
    List,
    Mapping,
    NewType,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

import graphql

from apischema import settings
from apischema.aliases import Aliaser
from apischema.conversions.conversions import AnyConversion, DefaultConversion
from apischema.conversions.visitor import (
    Conv,
    Deserialization,
    DeserializationVisitor,
    Serialization,
    SerializationVisitor,
)
from apischema.graphql.interfaces import get_interfaces, is_interface
from apischema.graphql.resolvers import (
    Resolver,
    get_resolvers,
    none_error_handler,
    partial_serialization_method_factory,
    resolver_parameters,
    resolver_resolve,
)
from apischema.json_schema.schema import get_field_schema, get_method_schema, get_schema
from apischema.metadata.keys import SCHEMA_METADATA
from apischema.objects import ObjectField
from apischema.objects.visitor import (
    DeserializationObjectVisitor,
    ObjectVisitor,
    SerializationObjectVisitor,
)
from apischema.ordering import Ordering, sort_by_order
from apischema.recursion import RecursiveConversionsVisitor
from apischema.schemas import Schema, merge_schema
from apischema.serialization import SerializationMethod, serialize
from apischema.serialization.serialized_methods import ErrorHandler
from apischema.type_names import TypeName, TypeNameFactory, get_type_name
from apischema.types import AnyType, NoneType, OrderedDict, Undefined, UndefinedType
from apischema.typing import get_args, get_origin, is_annotated
from apischema.utils import (
    Lazy,
    as_predicate,
    context_setter,
    deprecate_kwargs,
    empty_dict,
    get_args2,
    get_origin2,
    get_origin_or_type,
    identity,
    is_union_of,
    to_camel_case,
)

JsonScalar = graphql.GraphQLScalarType(
    "JSON",
    specified_by_url="http://www.ecma-international.org/publications/files/ECMA-ST/ECMA-404.pdf",
)
GRAPHQL_PRIMITIVE_TYPES = {
    int: graphql.GraphQLInt,
    float: graphql.GraphQLFloat,
    str: graphql.GraphQLString,
    bool: graphql.GraphQLBoolean,
}

ID = NewType("ID", str)


class MissingName(Exception):
    pass


class Nullable(Exception):
    pass


T = TypeVar("T")
Thunk = Union[Callable[[], T], T]

TypeThunk = Thunk[graphql.GraphQLType]


def exec_thunk(thunk: TypeThunk, *, non_null=None) -> Any:
    result = thunk if isinstance(thunk, graphql.GraphQLType) else thunk()
    if non_null is True and not isinstance(result, graphql.GraphQLNonNull):
        return graphql.GraphQLNonNull(result)  # type: ignore
    if non_null is False and isinstance(result, graphql.GraphQLNonNull):
        return result.of_type
    return result


def get_parameter_schema(
    func: Callable, parameter: Parameter, field: ObjectField
) -> Optional[Schema]:
    from apischema import settings

    return merge_schema(
        settings.base_schema.parameter(func, parameter, field.alias), field.schema
    )


def merged_schema(
    schema: Optional[Schema], tp: Optional[AnyType]
) -> Tuple[Optional[Schema], Mapping[str, Any]]:
    if is_annotated(tp):
        for annotation in reversed(get_args(tp)[1:]):
            if isinstance(annotation, TypeNameFactory):
                break
            elif isinstance(annotation, Mapping) and SCHEMA_METADATA in annotation:
                schema = merge_schema(annotation[SCHEMA_METADATA], schema)
    schema_dict: Dict[str, Any] = {}
    if schema is not None:
        schema.merge_into(schema_dict)
    return schema, schema_dict


def get_description(
    schema: Optional[Schema], tp: Optional[AnyType] = None
) -> Optional[str]:
    _, schema_dict = merged_schema(schema, tp)
    return schema_dict.get("description")


def get_deprecated(
    schema: Optional[Schema], tp: Optional[AnyType] = None
) -> Optional[str]:
    schema, schema_dict = merged_schema(schema, tp)
    if not schema_dict.get("deprecated", False):
        return None
    while schema is not None:
        if schema.annotations is not None:
            if isinstance(schema.annotations.deprecated, str):
                return schema.annotations.deprecated
            elif schema.annotations.deprecated:
                return graphql.DEFAULT_DEPRECATION_REASON
        schema = schema.child
    return graphql.DEFAULT_DEPRECATION_REASON


@dataclass(frozen=True)
class ResolverField:
    resolver: Resolver
    types: Mapping[str, AnyType]
    parameters: Sequence[Parameter]
    metadata: Mapping[str, Mapping]
    subscribe: Optional[Callable] = None


IdPredicate = Callable[[AnyType], bool]
UnionNameFactory = Callable[[Sequence[str]], str]


GraphQLTp = TypeVar("GraphQLTp", graphql.GraphQLInputType, graphql.GraphQLOutputType)

FactoryFunction = Callable[[Optional[str], Optional[str]], GraphQLTp]


@dataclass(frozen=True)
class TypeFactory(Generic[GraphQLTp]):
    factory: FactoryFunction[GraphQLTp]
    name: Optional[str] = None
    description: Optional[str] = None
    # non_null cannot be a field because it can not be forward to factories called in
    # wrapping factories (e.g. recursive wrapper)

    def merge(
        self, type_name: TypeName = TypeName(), schema: Optional[Schema] = None
    ) -> "TypeFactory[GraphQLTp]":
        if type_name == TypeName() and schema is None:
            return self
        return replace(
            self,
            name=type_name.graphql or self.name,
            description=get_description(schema) or self.description,
        )

    @property
    def type(self) -> GraphQLTp:
        return self.factory(self.name, self.description)  # type: ignore

    @property
    def raw_type(self) -> GraphQLTp:
        tp = self.type
        return tp.of_type if isinstance(tp, graphql.GraphQLNonNull) else tp


def unwrap_name(name: Optional[str], tp: AnyType) -> str:
    if name is None:
        raise TypeError(f"Missing name for {tp}")
    return name


Method = TypeVar("Method", bound=Callable[..., TypeFactory])


def cache_type(method: Method) -> Method:
    @wraps(method)
    def wrapper(self: "SchemaBuilder", *args, **kwargs):
        factory = method(self, *args, **kwargs)

        @wraps(factory.factory)  # type: ignore
        def name_cache(
            name: Optional[str], description: Optional[str]
        ) -> graphql.GraphQLNonNull:
            if name is None:
                return graphql.GraphQLNonNull(factory.factory(name, description))  # type: ignore
            # Method is in cache key because scalar types will have the same method,
            # and then be shared by both visitors, while input/output types will have
            # their own cache entry.
            if (name, method, description) in self._cache_by_name:
                tp, cached_args = self._cache_by_name[(name, method, description)]
                if cached_args == (args, kwargs):
                    return tp
            tp = graphql.GraphQLNonNull(factory.factory(name, description))  # type: ignore
            # Don't put args in cache in order to avoid hashable issue
            self._cache_by_name[(name, method, description)] = (tp, (args, kwargs))
            return tp

        return replace(factory, factory=name_cache)

    return cast(Method, wrapper)


class SchemaBuilder(
    RecursiveConversionsVisitor[Conv, TypeFactory[GraphQLTp]],
    ObjectVisitor[TypeFactory[GraphQLTp]],
):
    types: Tuple[Type[graphql.GraphQLType], ...]

    def __init__(
        self,
        aliaser: Aliaser,
        enum_aliaser: Aliaser,
        enum_schemas: Mapping[Enum, Schema],
        default_conversion: DefaultConversion,
        id_type: graphql.GraphQLScalarType,
        is_id: Optional[IdPredicate],
    ):
        super().__init__(default_conversion)
        self.aliaser = aliaser
        self.enum_aliaser = enum_aliaser
        self.enum_schemas = enum_schemas
        self.id_type = id_type
        self.is_id = is_id or (lambda t: False)
        self._cache_by_name: Dict[
            Tuple[str, Callable, Optional[str]],
            Tuple[graphql.GraphQLNonNull, Tuple[tuple, dict]],
        ] = {}

    def _recursive_result(
        self, lazy: Lazy[TypeFactory[GraphQLTp]]
    ) -> TypeFactory[GraphQLTp]:
        def factory(name: Optional[str], description: Optional[str]) -> GraphQLTp:
            cached_fact = lazy()
            return cached_fact.factory(  # type: ignore
                name or cached_fact.name, description or cached_fact.description
            )

        return TypeFactory(factory)

    def annotated(
        self, tp: AnyType, annotations: Sequence[Any]
    ) -> TypeFactory[GraphQLTp]:
        factory = super().annotated(tp, annotations)
        type_name = False
        for annotation in reversed(annotations):
            if isinstance(annotation, TypeNameFactory):
                if type_name:
                    break
                type_name = True
                factory = factory.merge(annotation.to_type_name(tp))
            if isinstance(annotation, Mapping):
                if type_name:
                    factory = factory.merge(schema=annotation.get(SCHEMA_METADATA))
        return factory  # type: ignore

    @cache_type
    def any(self) -> TypeFactory[GraphQLTp]:
        def factory(
            name: Optional[str], description: Optional[str]
        ) -> graphql.GraphQLScalarType:
            if name is None:
                return JsonScalar
            else:
                return graphql.GraphQLScalarType(name, description=description)

        return TypeFactory(factory)

    @cache_type
    def collection(
        self, cls: Type[Collection], value_type: AnyType
    ) -> TypeFactory[GraphQLTp]:
        return TypeFactory(lambda *_: graphql.GraphQLList(self.visit(value_type).type))

    @cache_type
    def enum(self, cls: Type[Enum]) -> TypeFactory[GraphQLTp]:
        def factory(
            name: Optional[str], description: Optional[str]
        ) -> graphql.GraphQLEnumType:
            return graphql.GraphQLEnumType(
                unwrap_name(name, cls),
                {
                    self.enum_aliaser(name): graphql.GraphQLEnumValue(
                        member,
                        get_description(self.enum_schemas.get(member)),
                        get_deprecated(self.enum_schemas.get(member)),
                    )
                    for name, member in cls.__members__.items()
                },
                description=description,
            )

        return TypeFactory(factory)

    @cache_type
    def literal(self, values: Sequence[Any]) -> TypeFactory[GraphQLTp]:
        from apischema.typing import Literal

        if not all(isinstance(v, str) for v in values):
            raise TypeError("apischema GraphQL only support Literal of strings")

        def factory(
            name: Optional[str], description: Optional[str]
        ) -> graphql.GraphQLEnumType:
            return graphql.GraphQLEnumType(
                unwrap_name(name, Literal[tuple(values)]),  # type: ignore
                dict(zip(map(self.enum_aliaser, values), values)),
                description=description,
            )

        return TypeFactory(factory)

    @cache_type
    def mapping(
        self, cls: Type[Mapping], key_type: AnyType, value_type: AnyType
    ) -> TypeFactory[GraphQLTp]:
        def factory(
            name: Optional[str], description: Optional[str]
        ) -> graphql.GraphQLScalarType:
            if name is not None:
                return graphql.GraphQLScalarType(name, description=description)
            else:
                return JsonScalar

        return TypeFactory(factory)

    def object(
        self, tp: AnyType, fields: Sequence[ObjectField]
    ) -> TypeFactory[GraphQLTp]:
        raise NotImplementedError

    @cache_type
    def primitive(self, cls: Type) -> TypeFactory[GraphQLTp]:
        def factory(
            name: Optional[str], description: Optional[str]
        ) -> graphql.GraphQLScalarType:
            assert cls is not NoneType
            if name is not None:
                return graphql.GraphQLScalarType(name, description=description)
            else:
                return GRAPHQL_PRIMITIVE_TYPES[cls]

        return TypeFactory(factory)

    def tuple(self, types: Sequence[AnyType]) -> TypeFactory[GraphQLTp]:
        raise TypeError("Tuple are not supported")

    def union(self, alternatives: Sequence[AnyType]) -> TypeFactory[GraphQLTp]:
        factories = self._union_results(
            (alt for alt in alternatives if alt is not NoneType)
        )
        if len(factories) == 1:
            factory = factories[0]
        else:
            factory = self._visited_union(factories)
        if NoneType in alternatives or UndefinedType in alternatives:

            def nullable(name: Optional[str], description: Optional[str]) -> GraphQLTp:
                res = factory.factory(name, description)  # type: ignore
                return res.of_type if isinstance(res, graphql.GraphQLNonNull) else res

            return replace(factory, factory=nullable)
        else:
            return factory

    def visit_conversion(
        self,
        tp: AnyType,
        conversion: Optional[Conv],
        dynamic: bool,
        next_conversion: Optional[AnyConversion] = None,
    ) -> TypeFactory[GraphQLTp]:
        if not dynamic and self.is_id(tp) or tp == ID:
            return TypeFactory(lambda *_: graphql.GraphQLNonNull(self.id_type))
        factory = super().visit_conversion(tp, conversion, dynamic, next_conversion)
        if not dynamic:
            factory = factory.merge(get_type_name(tp), get_schema(tp))
            if get_args(tp):
                factory = factory.merge(schema=get_schema(get_origin(tp)))
        return factory  # type: ignore


FieldType = TypeVar("FieldType", graphql.GraphQLInputField, graphql.GraphQLField)


class BaseField(Generic[FieldType]):
    name: str
    ordering: Optional[Ordering]

    def items(self) -> Iterable[Tuple[str, FieldType]]:
        raise NotImplementedError


@dataclass
class NormalField(BaseField[FieldType]):
    alias: str
    name: str
    field: Lazy[FieldType]
    ordering: Optional[Ordering]

    def items(self) -> Iterable[Tuple[str, FieldType]]:
        yield self.alias, self.field()


@dataclass
class FlattenedField(BaseField[FieldType]):
    name: str
    ordering: Optional[Ordering]
    type: TypeFactory

    def items(self) -> Iterable[Tuple[str, FieldType]]:
        tp = self.type.raw_type
        if not isinstance(
            tp,
            (
                graphql.GraphQLObjectType,
                graphql.GraphQLInterfaceType,
                graphql.GraphQLInputObjectType,
            ),
        ):
            raise FlattenedError(self)
        yield from tp.fields.items()


class FlattenedError(Exception):
    def __init__(self, field: FlattenedField):
        self.field = field


def merge_fields(cls: type, fields: Sequence[BaseField]) -> Dict[str, FieldType]:
    try:
        sorted_fields = sort_by_order(
            cls, fields, lambda f: f.name, lambda f: f.ordering
        )
    except FlattenedError as err:
        raise TypeError(
            f"Flattened field {cls.__name__}.{err.field.name}"
            f" must have an object type"
        )
    return OrderedDict(chain.from_iterable(map(lambda f: f.items(), sorted_fields)))


class InputSchemaBuilder(
    SchemaBuilder[Deserialization, graphql.GraphQLInputType],
    DeserializationVisitor[TypeFactory[graphql.GraphQLInputType]],
    DeserializationObjectVisitor[TypeFactory[graphql.GraphQLInputType]],
):
    types = graphql.type.definition.graphql_input_types

    def _field(
        self, tp: AnyType, field: ObjectField
    ) -> Lazy[graphql.GraphQLInputField]:
        field_type = field.type
        field_default = graphql.Undefined if field.required else field.get_default()
        default: Any = graphql.Undefined
        # Don't put `null` default + handle Undefined as None
        if field_default in {None, Undefined}:
            field_type = Optional[field_type]
        elif field_default is not graphql.Undefined:
            try:
                default = serialize(
                    field_type,
                    field_default,
                    aliaser=self.aliaser,
                    conversion=field.deserialization,
                )
            except Exception:
                field_type = Optional[field_type]
        factory = self.visit_with_conv(field_type, field.deserialization)
        return lambda: graphql.GraphQLInputField(
            factory.type,  # type: ignore
            default_value=default,
            description=get_description(get_field_schema(tp, field), field.type),
        )

    @cache_type
    def object(
        self, tp: AnyType, fields: Sequence[ObjectField]
    ) -> TypeFactory[graphql.GraphQLInputType]:
        visited_fields: List[BaseField] = []
        for field in fields:
            if not field.is_aggregate:
                normal_field = NormalField(
                    self.aliaser(field.alias),
                    field.name,
                    self._field(tp, field),
                    field.ordering,
                )
                visited_fields.append(normal_field)
            elif field.flattened:
                flattened_fields = FlattenedField(
                    field.name,
                    field.ordering,
                    self.visit_with_conv(field.type, field.deserialization),
                )
                visited_fields.append(flattened_fields)

        def factory(
            name: Optional[str], description: Optional[str]
        ) -> graphql.GraphQLInputObjectType:
            name = unwrap_name(name, tp)
            if not name.endswith("Input"):
                name += "Input"
            return graphql.GraphQLInputObjectType(
                name,
                lambda: merge_fields(get_origin_or_type(tp), visited_fields),
                description,
            )

        return TypeFactory(factory)

    def _visited_union(
        self, results: Sequence[TypeFactory]
    ) -> TypeFactory[graphql.GraphQLInputType]:
        # Check must be done here too because _union_result is used by visit_conversion
        if len(results) != 1:
            raise TypeError("Union are not supported for input")
        return results[0]


Func = TypeVar("Func", bound=Callable)


class OutputSchemaBuilder(
    SchemaBuilder[Serialization, graphql.GraphQLOutputType],
    SerializationVisitor[TypeFactory[graphql.GraphQLOutputType]],
    SerializationObjectVisitor[TypeFactory[graphql.GraphQLOutputType]],
):
    types = graphql.type.definition.graphql_output_types

    def __init__(
        self,
        aliaser: Aliaser,
        enum_aliaser: Aliaser,
        enum_schemas: Mapping[Enum, Schema],
        default_conversion: DefaultConversion,
        id_type: graphql.GraphQLScalarType,
        is_id: Optional[IdPredicate],
        union_name_factory: UnionNameFactory,
        default_deserialization: DefaultConversion,
    ):
        super().__init__(
            aliaser, enum_aliaser, enum_schemas, default_conversion, id_type, is_id
        )
        self.union_name_factory = union_name_factory
        self.input_builder = InputSchemaBuilder(
            self.aliaser,
            self.enum_aliaser,
            self.enum_schemas,
            default_deserialization,
            self.id_type,
            self.is_id,
        )
        # Share the same cache for input_builder in order to share scalar types
        self.input_builder._cache_by_name = self._cache_by_name
        self.get_flattened: Optional[Callable[[Any], Any]] = None

    def _field_serialization_method(self, field: ObjectField) -> SerializationMethod:
        return partial_serialization_method_factory(
            self.aliaser, field.serialization, self.default_conversion
        )(Optional[field.type] if field.none_as_undefined else field.type)

    def _wrap_resolve(self, resolve: Func) -> Func:
        if self.get_flattened is None:
            return resolve
        else:
            get_flattened = self.get_flattened

            def resolve_wrapper(__obj, __info, **kwargs):
                return resolve(get_flattened(__obj), __info, **kwargs)

            return cast(Func, resolve_wrapper)

    def _field(self, tp: AnyType, field: ObjectField) -> Lazy[graphql.GraphQLField]:
        field_name = field.name
        partial_serialize = self._field_serialization_method(field)

        @self._wrap_resolve
        def resolve(obj, _):
            return partial_serialize(getattr(obj, field_name))

        factory = self.visit_with_conv(field.type, field.serialization)
        field_schema = get_field_schema(tp, field)
        return lambda: graphql.GraphQLField(
            factory.type,
            None,
            resolve,
            description=get_description(field_schema, field.type),
            deprecation_reason=get_deprecated(field_schema, field.type),
        )

    def _resolver(
        self, tp: AnyType, field: ResolverField
    ) -> Lazy[graphql.GraphQLField]:
        resolve = self._wrap_resolve(
            resolver_resolve(
                field.resolver,
                field.types,
                self.aliaser,
                self.input_builder.default_conversion,
                self.default_conversion,
            )
        )
        args = None
        if field.parameters is not None:
            args = {}
            for param in field.parameters:
                default: Any = graphql.Undefined
                param_type = field.types[param.name]
                if is_union_of(param_type, graphql.GraphQLResolveInfo):
                    break
                param_field = ObjectField(
                    param.name,
                    param_type,
                    param.default is Parameter.empty,
                    field.metadata.get(param.name, empty_dict),
                    default=param.default,
                )
                if param_field.required:
                    pass
                # Don't put `null` default + handle Undefined as None
                # also https://github.com/python/typing/issues/775
                elif param.default in {None, Undefined}:
                    param_type = Optional[param_type]
                # param.default == graphql.Undefined means the parameter is required
                # even if it has a default
                elif param.default not in {Parameter.empty, graphql.Undefined}:
                    try:
                        default = serialize(
                            param_type,
                            param.default,
                            fall_back_on_any=False,
                            check_type=True,
                        )
                    except Exception:
                        param_type = Optional[param_type]
                arg_factory = self.input_builder.visit_with_conv(
                    param_type, param_field.deserialization
                )
                description = get_description(
                    get_parameter_schema(field.resolver.func, param, param_field),
                    param_field.type,
                )

                def arg_thunk(
                    arg_factory=arg_factory, default=default, description=description
                ) -> graphql.GraphQLArgument:
                    return graphql.GraphQLArgument(
                        arg_factory.type, default, description
                    )

                args[self.aliaser(param_field.alias)] = arg_thunk
        factory = self.visit_with_conv(field.types["return"], field.resolver.conversion)
        field_schema = get_method_schema(tp, field.resolver)
        return lambda: graphql.GraphQLField(
            factory.type,  # type: ignore
            {name: arg() for name, arg in args.items()} if args else None,
            resolve,
            field.subscribe,
            get_description(field_schema),
            get_deprecated(field_schema),
        )

    def _visit_flattened(
        self, field: ObjectField
    ) -> TypeFactory[graphql.GraphQLOutputType]:
        get_prev_flattened = (
            self.get_flattened if self.get_flattened is not None else identity
        )
        field_name = field.name
        partial_serialize = self._field_serialization_method(field)

        def get_flattened(obj):
            return partial_serialize(getattr(get_prev_flattened(obj), field_name))

        with context_setter(self):
            self.get_flattened = get_flattened
            return self.visit_with_conv(field.type, field.serialization)

    @cache_type
    def object(
        self,
        tp: AnyType,
        fields: Sequence[ObjectField],
        resolvers: Sequence[ResolverField] = (),
    ) -> TypeFactory[graphql.GraphQLOutputType]:
        cls = get_origin_or_type(tp)
        visited_fields: List[BaseField[graphql.GraphQLField]] = []
        flattened_factories = []
        for field in fields:
            if not field.is_aggregate:
                normal_field = NormalField(
                    self.aliaser(field.name),
                    field.name,
                    self._field(tp, field),
                    field.ordering,
                )
                visited_fields.append(normal_field)
            elif field.flattened:
                flattened_factory = self._visit_flattened(field)
                flattened_factories.append(flattened_factory)
                visited_fields.append(
                    FlattenedField(field.name, field.ordering, flattened_factory)
                )
        resolvers = list(resolvers)
        for resolver, types in get_resolvers(tp):
            resolver_field = ResolverField(
                resolver, types, resolver.parameters, resolver.parameters_metadata
            )
            resolvers.append(resolver_field)
        for resolver_field in resolvers:
            normal_field = NormalField(
                self.aliaser(resolver_field.resolver.alias),
                resolver_field.resolver.func.__name__,
                self._resolver(tp, resolver_field),
                resolver_field.resolver.ordering,
            )
            visited_fields.append(normal_field)

        interface_thunk = None
        interfaces = list(map(self.visit, get_interfaces(cls)))
        if interfaces or flattened_factories:

            def interface_thunk() -> Collection[graphql.GraphQLInterfaceType]:
                all_interfaces = {
                    cast(graphql.GraphQLInterfaceType, i.raw_type) for i in interfaces
                }
                for flattened_factory in flattened_factories:
                    flattened = cast(
                        Union[graphql.GraphQLObjectType, graphql.GraphQLInterfaceType],
                        flattened_factory.raw_type,
                    )
                    if isinstance(flattened, graphql.GraphQLObjectType):
                        all_interfaces.update(flattened.interfaces)
                    elif isinstance(flattened, graphql.GraphQLInterfaceType):
                        all_interfaces.add(flattened)
                return sorted(all_interfaces, key=lambda i: i.name)

        def factory(
            name: Optional[str], description: Optional[str]
        ) -> Union[graphql.GraphQLObjectType, graphql.GraphQLInterfaceType]:
            name = unwrap_name(name, cls)
            if is_interface(cls):
                return graphql.GraphQLInterfaceType(
                    name,
                    lambda: merge_fields(cls, visited_fields),
                    interface_thunk,
                    description=description,
                )
            else:
                return graphql.GraphQLObjectType(
                    name,
                    lambda: merge_fields(cls, visited_fields),
                    interface_thunk,
                    is_type_of=lambda obj, _: isinstance(obj, cls),
                    description=description,
                )

        return TypeFactory(factory)

    def typed_dict(
        self, tp: Type, types: Mapping[str, AnyType], required_keys: Collection[str]
    ) -> TypeFactory[graphql.GraphQLOutputType]:
        raise TypeError("TypedDict are not supported in output schema")

    @cache_type
    def _visited_union(
        self, results: Sequence[TypeFactory]
    ) -> TypeFactory[graphql.GraphQLOutputType]:
        def factory(
            name: Optional[str], description: Optional[str]
        ) -> graphql.GraphQLOutputType:
            types = [factory.raw_type for factory in results]
            if name is None:
                name = self.union_name_factory([t.name for t in types])
            return graphql.GraphQLUnionType(name, types, description=description)

        return TypeFactory(factory)


async_iterable_origins = set(map(get_origin, (AsyncIterable[Any], AsyncIterator[Any])))

_fake_type = cast(type, ...)


@dataclass(frozen=True)
class Operation(Generic[T]):
    function: Callable[..., T]
    alias: Optional[str] = None
    conversion: Optional[AnyConversion] = None
    error_handler: ErrorHandler = Undefined
    order: Optional[Ordering] = None
    schema: Optional[Schema] = None
    parameters_metadata: Mapping[str, Mapping] = field_(default_factory=dict)


class Query(Operation):
    pass


class Mutation(Operation):
    pass


@dataclass(frozen=True)
class Subscription(Operation[AsyncIterable]):
    resolver: Optional[Callable] = None


Op = TypeVar("Op", bound=Operation)


def operation_resolver(operation: Union[Callable, Op], op_class: Type[Op]) -> Resolver:
    if not isinstance(operation, op_class):
        operation = op_class(operation)  # type: ignore
    error_handler: Optional[Callable]
    if operation.error_handler is Undefined:
        error_handler = None
    elif operation.error_handler is None:
        error_handler = none_error_handler
    else:
        error_handler = operation.error_handler
    op = operation.function
    if iscoroutinefunction(op):

        async def wrapper(_, *args, **kwargs):
            return await op(*args, **kwargs)

    else:

        def wrapper(_, *args, **kwargs):
            return op(*args, **kwargs)

    wrapper.__annotations__ = op.__annotations__

    (*parameters,) = resolver_parameters(operation.function, check_first=True)
    return Resolver(
        wrapper,
        operation.alias or operation.function.__name__,
        operation.conversion,
        error_handler,
        operation.order,
        operation.schema,
        parameters,
        operation.parameters_metadata,
    )


@deprecate_kwargs({"union_ref": "union_name"})
def graphql_schema(
    *,
    query: Iterable[Union[Callable, Query]] = (),
    mutation: Iterable[Union[Callable, Mutation]] = (),
    subscription: Iterable[Union[Callable[..., AsyncIterable], Subscription]] = (),
    types: Iterable[Type] = (),
    directives: Optional[Collection[graphql.GraphQLDirective]] = None,
    description: Optional[str] = None,
    extensions: Optional[Dict[str, Any]] = None,
    aliaser: Optional[Aliaser] = to_camel_case,
    enum_aliaser: Optional[Aliaser] = str.upper,
    enum_schemas: Optional[Mapping[Enum, Schema]] = None,
    id_types: Union[Collection[AnyType], IdPredicate] = (),
    id_encoding: Tuple[
        Optional[Callable[[str], Any]], Optional[Callable[[Any], str]]
    ] = (None, None),
    union_name: UnionNameFactory = "Or".join,
    default_deserialization: DefaultConversion = None,
    default_serialization: DefaultConversion = None,
) -> graphql.GraphQLSchema:
    if aliaser is None:
        aliaser = settings.aliaser
    if enum_aliaser is None:
        enum_aliaser = lambda s: s
    if default_deserialization is None:
        default_deserialization = settings.deserialization.default_conversion
    if default_serialization is None:
        default_serialization = settings.serialization.default_conversion
    query_fields: List[ResolverField] = []
    mutation_fields: List[ResolverField] = []
    subscription_fields: List[ResolverField] = []
    for operations, op_class, fields in [
        (query, Query, query_fields),
        (mutation, Mutation, mutation_fields),
    ]:
        for operation in operations:  # type: ignore
            resolver = operation_resolver(operation, op_class)
            resolver_field = ResolverField(
                resolver,
                resolver.types(),
                resolver.parameters,
                resolver.parameters_metadata,
            )
            fields.append(resolver_field)
    for sub_op in subscription:  # type: ignore
        if not isinstance(sub_op, Subscription):
            sub_op = Subscription(sub_op)  # type: ignore
        sub_parameters: Sequence[Parameter]
        if sub_op.resolver is not None:
            subscriber2 = operation_resolver(sub_op, Subscription)
            _, *sub_parameters = resolver_parameters(sub_op.resolver, check_first=False)
            resolver = Resolver(
                sub_op.resolver,
                sub_op.alias or sub_op.resolver.__name__,
                sub_op.conversion,
                subscriber2.error_handler,
                sub_op.order,
                sub_op.schema,
                sub_parameters,
                sub_op.parameters_metadata,
            )
            sub_types = resolver.types()
            subscriber = replace(subscriber2, error_handler=None)
            subscribe = resolver_resolve(
                subscriber,
                subscriber.types(),
                aliaser,
                default_deserialization,
                default_serialization,
                serialized=False,
            )
        else:
            subscriber2 = operation_resolver(sub_op, Subscription)
            resolver = Resolver(
                lambda _: _,
                subscriber2.alias,
                sub_op.conversion,
                subscriber2.error_handler,
                sub_op.order,
                sub_op.schema,
                (),
                {},
            )
            subscriber = replace(subscriber2, error_handler=None)
            sub_parameters = subscriber.parameters
            sub_types = subscriber.types()
            if get_origin2(sub_types["return"]) not in async_iterable_origins:
                raise TypeError(
                    "Subscriptions must return an AsyncIterable/AsyncIterator"
                )
            event_type = get_args2(sub_types["return"])[0]
            subscribe = resolver_resolve(
                subscriber,
                sub_types,
                aliaser,
                default_deserialization,
                default_serialization,
                serialized=False,
            )
            sub_types = {**sub_types, "return": resolver.return_type(event_type)}

        resolver_field = ResolverField(
            resolver, sub_types, sub_parameters, sub_op.parameters_metadata, subscribe
        )
        subscription_fields.append(resolver_field)

    is_id = as_predicate(id_types)
    if id_encoding == (None, None):
        id_type: graphql.GraphQLScalarType = graphql.GraphQLID
    else:
        id_deserializer, id_serializer = id_encoding
        id_type = graphql.GraphQLScalarType(
            name="ID",
            serialize=id_serializer or graphql.GraphQLID.serialize,
            parse_value=id_deserializer or graphql.GraphQLID.parse_value,
            parse_literal=graphql.GraphQLID.parse_literal,
            description=graphql.GraphQLID.description,
        )

    output_builder = OutputSchemaBuilder(
        aliaser,
        enum_aliaser,
        enum_schemas or {},
        default_serialization,
        id_type,
        is_id,
        union_name,
        default_deserialization,
    )

    def root_type(
        name: str, fields: Sequence[ResolverField]
    ) -> Optional[graphql.GraphQLObjectType]:
        if not fields:
            return None
        tp, type_name = type(name, (), {}), TypeName(graphql=name)
        return output_builder.object(tp, (), fields).merge(type_name, None).raw_type  # type: ignore

    return graphql.GraphQLSchema(
        query=root_type("Query", query_fields),
        mutation=root_type("Mutation", mutation_fields),
        subscription=root_type("Subscription", subscription_fields),
        types=[output_builder.visit(cls).raw_type for cls in types],  # type: ignore
        directives=directives,
        description=description,
        extensions=extensions,
    )



__all__ = ["NotNull", "Skip"]
from typing import TypeVar, Union

from apischema.visitor import Unsupported

Skip = Unsupported

T = TypeVar("T")
try:
    from apischema.typing import Annotated

    NotNull = Union[T, Annotated[None, Skip]]
except ImportError:

    class _NotNull:
        def __getitem__(self, item):
            raise TypeError("NotNull requires Annotated (PEP 593)")

    NotNull = _NotNull()  # type: ignore
